package NoRFC;
use strict;
use utf8;
use feature qw/ say /;

=encoding utf8

=head1 NAME

NoRFC - Management library of noRFC documents

=head1 VERSION

Version 1.08

=head1 SYNOPSIS

    use NoRFC;

=head1 ABSTRACT

norfc - program which performs management of noRFC documents

=head1 DESCRIPTION

This module provides norfc functionality

=head2 again

Internal use only

=head2 categories

Returns list of available categories

=head2 client

    my $client = $app->client;

Returns the Mojo client (user agent) instance

=head2 decrypt_file

    my $output = $app->decrypt_file($file_enc, $file_dec);
        die($app->error) if $app->error;

Decrypt file and returns GPG output data

=head2 encrypt_file

    my $output = $app->encrypt_file($fingerprint, $file_dec, $file_enc);
        die($app->error) if $app->error;

Encrypt file and returns GPG output data

=head2 extract

    my %extract_data = $app->extract(
            num => "NoRFC-R200000116",
            pub => 1614716051,
            category => "Etcetera",
            author => "Anonymous",
            title => "Test document",
            content => "Content",
            signature => "...xxx...",
            dir => "./",
        );

This method performs extraction the noRFC document and detached signature by number of specified NoRFC
Will be used local database directory as directory (dir) by default

Returns structure

    (
        docfile => "/full/path/to/document.txt",
        sigfile => "/full/path/to/document.txt.sig",
    )

 or error otherwise:

    die $app->error unless %extract_data;

=head2 fingerprint

Returns current fingerprint

=head2 getAuthor

    my $author = $app->getAuthor;

Sets the current fingerprint and returns the author name

=head2 getCategory

    my $category = $app->getCategory;
    my $category = $app->getCategory( "dafault" );

Returns category of the noRFC document

=head2 getStatus

    my $status = $app->getStatus;

Returns status of the noRFC document

=head2 getTitle

    my $title = $app->getTitle;
    my $title = $app->getTitle( "dafault" );

Returns title of the noRFC document

=head2 getType

    my $type = $app->getType;

Returns type of the noRFC document

=head2 gnupg

    my $gnupg = $app->gnupg();

Returns gnupg object

=head2 is_valid

    $app->is_valid("NoRFC-R200000116") or die ($self->error);

Returns 0 and sets error if specified document number is not valid,
and returns 1 if is valid

    $app->is_valid("NoRFC-R200000116", QUIET_MODE) or die ("Incorrect document number");

Returns 0 in quiet mode (without error setting) if specified document number is not valid,
and returns 1 if is valid

=head2 private_keys_info

    my %users = $app->private_keys_info();

Returns structure that describes private keys in format:

    fingerprint => {name, validity, order}

=head2 raise

    $app->raise("Error string");

Prints error string to STDERR pipe

=head2 relations

    my %relations = $app->relations;

Returns relations from command-line options - "Obsoletes" and "Updates"

=head2 sign

    my $sign = $app->sign( $fingerprint, "Document body" )
        or die($app->error());

Signs document and returns detached signature

=head2 sign_file

    my $output = $app->sign_file($fingerprint, $file, $sig_file);
        die($app->error) if $app->error;

Signs file to detached signature file and returns GPG output data

=head2 slurp_io

    my $data = $app->slurp_io(); # STDIO pipe
    my $data = $app->slurp_io($io); # IO::Handle and etc. pipes

Reads and returns all bytes from IO pipe. Default: STDIO pipe

=head2 store

    my $store = $app->store();

Returns store object

=head2 verify_file

    $app->verify_file($norfc_sig_file, $norfc_file) or die "Bad!"
    $app->verify_file($norfc_sig_file, $norfc_file, QUIET_MODE) or die "Bad!"

Verifies document and returns status. The QUIET_MODE flag suppress gpg pipes data

=head2 yaml_load

    my $data = $app->yaml_load( "file.yml" );
    die $self->error if $self->error;

Load YAML data from file

=head2 yaml_save

    $app->yaml_save( "file.yml", {foo => 123} )
        or die $self->error;

Save YAML data to file

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<Mojolicious>, L<GnuPG::Handles>, L<GnuPG::Interface>, L<YAML::XS>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/$VERSION/;
$VERSION = '1.08';

use base qw/CTK::App/;

use Encode; # Encode::_utf8_on();
use Encode::Locale;

use Cwd;
use File::stat qw//;
use File::Basename qw/basename/;
use File::Spec;
use File::Find qw/find/;
use File::Path qw/rmtree/;
use YAML::XS qw/DumpFile LoadFile/;
use Try::Tiny;
use IO::Handle;
use GnuPG::Handles;
use GnuPG::Interface;
use Text::SimpleTable;
use POSIX qw//;

use Mojo::File qw/path/;
use Mojo::Util qw/steady_time/;
use Mojo::Server::Prefork;

use CTK::Util qw/ rundir trim dtf tz_diff variant_stf preparedir dformat bload bsave isTrueFlag which /;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;

use NoRFC::Const qw/ :GENERAL :MAPS :SERVER :NUMBER /;
use NoRFC::Util qw/ :DEFAULT explain m11r_checkdigit parsewords fbytes fdate filesize md5sum sha256sum hash2attr attr2hash /;
use NoRFC::Number;
use NoRFC::Document;
use NoRFC::Store;
use NoRFC::Server;
use NoRFC::Client;
use NoRFC::Progress;

use constant NORFC_USAGE => <<'USAGE';

Show general information:
    norfc status
Show list of existing documents:
    norfc search
Check the document:
    norfc check <NORFC>
Show document information:
    norfc show <NORFC>
Get the noRFC document:
    norfc get <NORFC>
Generation new noRFC document:
    norfc gen <FILE>
Update document:
    norfc set <NORFC> <FILE>
Delete document:
    norfc del <NORFC>
Add relation:
    norfc rel add <NORFC> -O|-U <EXT_NORFC_1>
Get list of relations:
    norfc rel get <NORFC> [-O|-U <EXT_NORFC_1>]
Delete relation:
    norfc rel del <NORFC> [-O|-U <EXT_NORFC_1>]
Export all documents from database to file archive:
    norfc export [<ARCHIVE.tgz>]
Import all documents from file archive to database:
    norfc import <ARCHIVE.tgz>
Show full documentation:
    norfc -H

USAGE

__PACKAGE__->register_handler(
    handler     => "version",
    description => "Project version",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    printf("%s/%s\n", PROJECTNAME, $self->VERSION);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "usage",
    description => "Project usage page",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    print cyan(NORFC_USAGE);
    return 1;
});

__PACKAGE__->register_handler(
    handler     => "status",
    description => "Get status information",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    printf("Version             : %s\n", $self->VERSION);
    printf("Data dir            : %s\n", $self->datadir);
    printf("Status file         : %s\n", File::Spec->catfile($self->datadir, YFILE));

    # Configuration status
    printf("Config file         : %s\n", $self->configfile);
    printf("Config status       : %s\n", $self->conf("loadstatus") ? green("OK") : magenta("ERROR: not loaded"));
        $self->raise($self->configobj->error) if !$self->configobj->status and length($self->configobj->error);
    if ($self->conf("loadstatus")) {
        printf("Configuration       : %s\n", $self->verbosemode ? "" : yellow("please use -v flag to show it"));
        print explain($self->config) if $self->verbosemode;
    }

    # Fingerprint
    my $fingerprint = lvalue($self->conf("fingerprint"));
    if ($fingerprint) {
        if (_fingerprint_test($fingerprint)) {
            printf("Fingerprint         : %s\n", $fingerprint);
        } else {
            $self->raise("The fingerprint is incorrect");
        }
    }

    # Database status
    printf("Local DB status     :  %s\n", $store->status ? green("OK") : magenta("ERROR"));
    printf(" DB DSN             : %s\n", $store->dsn);
    $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;
    printf(" Documents          : %s\n", $store->status ? $store->count : yellow("unknown"));
    if ($store->status && $store->is_sqlite && $store->{file} && -e $store->{file}) {
        my $s = File::stat::stat($store->{file})->size;
        printf(" Filename           : %s\n", $store->{file});
        printf(" File size          : %s\n", sprintf("%s (%d bytes)", fbytes($s), $s));
        printf(" Modified           : %s\n", fdate(File::stat::stat($store->{file})->mtime || 0));
    }

    # Categories
    printf("Categories          : %s\n", join(", ", $self->categories(1)));
    printf("Extra Cats          : %s\n", join(", ", $self->categories(2)));

    # Server status
    my $client = $self->client;
    $client->check(); # Check
    printf("Server status       : %s\n", $client->status ? green("OK") : magenta($client->error));
    printf(" URL                : %s\n", $client->url ? $client->url->to_string : yellow("Unknown"));
    $self->debug($client->trace) unless $client->status;

    # Get info from server
    return 0 unless $client->status;
    $self->client->info() or do {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    };

    printf(" Version            : %s\n", $self->client->res->json("/version") // '');
    printf(" Datadir            : %s\n", $self->client->res->json("/datadir") // '');
    printf(" Message            : %s\n", $self->client->res->json("/message") // '');
    printf(" DB DSN             : %s\n", $self->client->res->json("/dsn") // '');
    printf(" Documents          : %d\n", $self->client->res->json("/total") // 0);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "categories",
    description => "Get categories",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    say "List of available noRFC document categories";
    say "";
    say cyan("NoRFC categories:\n\t%s", join("\n\t", $self->categories(1)));
    say "";
    say magenta("Extra categories:\n\t%s", join("\n\t", $self->categories(2)));

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "gen",
    description => "Generation a new noRFC document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my %attr;
    $attr{columns} = $self->option("width") if $self->option("width");
    $attr{tabstop} = $self->option("tabstop") if $self->option("tabstop");

    # Document type
    my $type = $self->getType;
    return 0 unless $type;

    # Status letter
    my $let = $self->getStatus;
    return 0 unless $let;
    $attr{encrypted} = 1 if $let eq 'X';

    # Category
    my $category = $self->getCategory;
    return 0 unless $category;

    # Title
    my $title = $self->getTitle;
    return 0 unless $title;

    # Author
    my $author = $self->getAuthor;
    return 0 unless $author;

    # Date data (defaults)
    my ($mday, $mon, $year, $yday) = (localtime(time))[3,4,5,7];
    $mon++; $yday++; $year += 1900;

    # Year
    if (my $y = $self->option("year")) {
        if ($y =~ /^[0-9]{4}$/) {
            $year = $y;
        } else {
            return $self->raise("Incorrect year! Please specify real year");
        }
    }

    # Year's day
    if (my $d = $self->option("day")) {
        if (is_int($d) && $d && $d <= 366) {
            $yday = $d;
        } else {
            return $self->raise("Incorrect year's day! Please specify real year's day (value between 1 and 365 or 366)");
        }
    }

    # Sequence number (counter)
    my $nn = 0;
    if ($nn = $self->option("number")) {
        unless (is_int($nn) && $nn && $nn <= NORFCNUM_NN_MAX) {
            return $self->raise("Incorrect number value! Please specify value between 1 and 999999");
        }
    }

    # Generate the new document number
    my $cli_status = $self->client->gendocnum(
            type    => $type,
            status  => $let,
            year    => $year,
            yday    => $yday,
            nn      => $nn,
        );
    unless ($cli_status) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }
    my $docnum = $self->client->res->json("/number") // "";
    #say explain($self->client->req->content);
    #$self->debug($self->client->res->body);

    # Source file
    my $file = shift(@arguments) // '';
    return $self->raise("No file of document specified") unless length($file);
    return $self->raise("File \"%s\" not found", $file) unless -e $file;
    $attr{binary} = 1 if -B $file; # Binary file?

    # Filenames (paths)
    my $file_raw = File::Spec->catfile($self->tempdir, sprintf("%s.raw", $docnum));
    my $file_enc = File::Spec->catfile($self->tempdir, sprintf("%s.enc", $docnum));
    my $file_txt = File::Spec->catfile($self->tempdir, sprintf("%s.txt", $docnum));
    my $file_sig = File::Spec->catfile($self->tempdir, sprintf("%s.txt.sig", $docnum));

    # Confirm
    my $pub = time();
    say magenta("Document number    : %s", $docnum);
    say magenta(" Type              : %s", $type);
    say magenta(" Status            : %s", STATUSES_MAP()->{$let} // "");
    say magenta(" Category          : %s", encode(locale => $category));
    say magenta(" Title             : %s", encode(locale => $title));
    say magenta(" Published         : %s", fdate($pub));
    my $updates = array($self->option("updates"));
    say magenta(" Updates           : %s", join(", ", @$updates)) if isnt_void($updates);
    my $obsoletes = array($self->option("obsoletes"));
    say magenta(" Obsoletes         : %s", join(", ", @$obsoletes)) if isnt_void($obsoletes);
    say magenta(" Source file       : %s", $file);
    say magenta("   File size       : %s", fbytes(filesize($file)));
    say magenta("   File md5        : %s", md5sum($file));
    say magenta("   File sha256     : %s", sha256sum($file));
    say magenta(" Fingerprint       : %s", $self->fingerprint);
    unless ($self->option("force")) {
        return $self->raise("Operation aborted") if $self->cli_prompt('All right?:','yes') !~ /^\s*y/i;
    }

    # Create document
    path($file)->copy_to($file_raw);
    my $doc = NoRFC::Document->new(
        columns     => $self->option("width"),
        tabstop     => $self->option("tabstop"),
        number      => $docnum,
        category    => $category,
        published   => $pub,
        author      => $author,
        updates     => $updates,
        obsoletes   => $obsoletes,
        title       => $title,
        file        => $file_raw,
    );

    # Save document
    $doc->save($file_txt) or return $self->raise(sprintf("Can't save file \"%s\"", $file_txt));

    # Sign file if fingerprint
    if ($self->fingerprint) {
        $self->sign_file($self->fingerprint, $file_txt, $file_sig);
        return 0 if $self->error;
    }

    # Encrypt file
    if ($attr{encrypted}) {
        return $self->raise("Skipped. No fingerprint found") unless $self->fingerprint;
        $self->encrypt_file($self->fingerprint, $file_raw, $file_enc);
        return 0 if $self->error;
    }

    # Send data to server
    $cli_status = $self->client->publish(
            number      => $docnum,
            published   => $pub,
            category    => $category,
            title       => $title,
            author      => $author,
            fileraw     => $attr{encrypted} ? $file_enc : $file_raw,
            filesig     => $file_sig,
            attributes  => hash2attr(%attr),
        );
    unless ($cli_status) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }
    #say explain($self->client->req->content);
    #$self->debug($self->client->res->body);
    #say magenta($self->client->res->json("/number"));
    #$self->debug(explain(\%attr));

    # Add relations
    my %relations = $self->relations;
    while (my ($ext, $name) = each %relations) {
        last unless $self->client->reladd($docnum, $ext, $name);
    }
    unless ($self->client->res->json("/status")) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }

    # Show result
    say "";
    say cyan("The document %s has been successfully saved to\n %s", $docnum, $file_txt);
    say cyan("For verification, use:\n gpg --verify %s", $file_sig);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "set",
    description => "Update the noRFC document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my %attr;
    $attr{columns} = $self->option("width") if $self->option("width");
    $attr{tabstop} = $self->option("tabstop") if $self->option("tabstop");

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;
    $attr{encrypted} = 1 if $norfcnum->status eq 'X';

    # Source file
    my $file = shift(@arguments) // '';
    return $self->raise("No file of document specified") unless length($file);
    return $self->raise("File not found") unless -e $file;

    # Get data from server
    my $cli_status = $self->client->info($docnum);
    unless ($cli_status) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }
    return $self->raise("Document not found in database") unless $self->client->res->json("/id");

    # Category, Title, Author
    my $category = $self->getCategory(decode(locale => $self->client->res->json("/category") // ''));
    my $title = $self->getTitle(decode(locale => $self->client->res->json("/title") // ''));
    my $author = $self->getAuthor;

    # Filenames (paths)
    my $file_raw = File::Spec->catfile($self->tempdir, sprintf("%s.raw", $docnum));
    my $file_enc = File::Spec->catfile($self->tempdir, sprintf("%s.enc", $docnum));
    my $file_txt = File::Spec->catfile($self->tempdir, sprintf("%s.txt", $docnum));
    my $file_sig = File::Spec->catfile($self->tempdir, sprintf("%s.txt.sig", $docnum));

    # Confirm
    my $pub = time();
    say magenta("Document number    : %s", $docnum);
    say magenta(" Id                : %s", $self->client->res->json("/id"));
    say magenta(" Category          : %s", encode(locale => $category));
    say magenta(" Title             : %s", encode(locale => $title));
    say magenta(" Published         : %s", fdate($pub));
    my $updates = array($self->option("updates"));
    say magenta(" Updates           : %s", join(", ", @$updates)) if isnt_void($updates);
    my $obsoletes = array($self->option("obsoletes"));
    say magenta(" Obsoletes         : %s", join(", ", @$obsoletes)) if isnt_void($obsoletes);
    say magenta(" Source file       : %s", $file);
    say magenta("   File size       : %s", fbytes(filesize($file)));
    say magenta("   File md5        : %s", md5sum($file));
    say magenta("   File sha256     : %s", sha256sum($file));
    say magenta(" Fingerprint       : %s", $self->fingerprint);
    unless ($self->option("force")) {
        return $self->raise("Operation aborted") if $self->cli_prompt('All right?:','yes') !~ /^\s*y/i;
    }

    # Create document
    path($file)->copy_to($file_raw);
    my $doc = NoRFC::Document->new(
        columns     => $self->option("width"),
        tabstop     => $self->option("tabstop"),
        number      => $docnum,
        category    => $category,
        published   => $pub,
        author      => $author,
        updates     => $updates,
        obsoletes   => $obsoletes,
        title       => $title,
        file        => $file_raw,
    );

    # Save document
    $attr{binary} = 1 if $doc->is_binary; # Binary file?
    $doc->save($file_txt) or return $self->raise(sprintf("Can't save file \"%s\"", $file_txt));

    # Sign file if fingerprint
    if ($self->fingerprint) {
        $self->sign_file($self->fingerprint, $file_txt, $file_sig);
        return 0 if $self->error;
    }

    # Encrypt file
    if ($attr{encrypted}) {
        return $self->raise("Skipped. No fingerprint found") unless $self->fingerprint;
        $self->encrypt_file($self->fingerprint, $file_raw, $file_enc);
        return 0 if $self->error;
    }

    # Send data to server
    $cli_status = $self->client->publish(
            number      => $docnum,
            published   => $pub,
            category    => $category,
            title       => $title,
            author      => $author,
            fileraw     => $attr{encrypted} ? $file_enc : $file_raw,
            filesig     => $file_sig,
            attributes  => hash2attr(%attr),
        );
    unless ($cli_status) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }
    #$self->debug(explain(\%attr));

    # Remove relations
    $self->client->reldel($docnum) or do {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    };

    # Add relations
    my %relations = $self->relations;
    while (my ($ext, $name) = each %relations) {
        last unless $self->client->reladd($docnum, $ext, $name);
    }
    unless ($self->client->res->json("/status")) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }

    # Show result
    say "";
    say cyan("The document %s has been successfully saved to\n %s", $docnum, $file_txt);
    say cyan("For verification, use:\n gpg --verify %s", $file_sig);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "get",
    description => "Extract noRFC document to current directory",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments)
        || $self->cli_prompt('Please provide the noRFC number (e.g., NoRFC-R000000000):'));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;

    # Extract document from server
    my %info = $self->extract($docnum, getcwd());
    return 0 unless %info;
    return $self->raise("Document not found") unless $info{id};

    # Show result
    say "";
    say cyan("The document %s has been successfully saved to\n %s", $docnum, $info{file_txt});
    say cyan("For verification, use:\n gpg --verify %s", $info{file_sig}) if -e $info{file_sig};

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "check",
    description => "Checking noRFC document",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $store = $self->store;

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments)
        || $self->cli_prompt('Please provide the noRFC number (e.g., NoRFC-R000000000):'));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;

    # Extract document from server
    my %info = $self->extract($docnum, $self->tempdir);
    return 0 unless %info;
    return $self->raise("Document not found") unless $info{id};

    # Found!
    say sprintf("Found document     : %s", $docnum);
    say sprintf(" Id                : %s", $info{id});
    say sprintf(" Category          : %s", encode(locale => $info{category} // ''));
    say sprintf(" Title             : %s", encode(locale => $info{title} // ''));
    say sprintf(" Published         : %s", $info{pub} ? fdate($info{pub}) : yellow("unknown"));
    say sprintf(" Updates           : %s", join(", ", @{($info{updates})})) if isnt_void($info{updates}) && is_array($info{updates});
    say sprintf(" Obsoletes         : %s", join(", ", @{($info{obsoletes})})) if isnt_void($info{obsoletes}) && is_array($info{obsoletes});
    say sprintf(" Author            : %s", encode(locale => $info{author} // ''));
    say sprintf(" Attributes        : %s", $info{attributes} // '');

    # Document file
    if ($info{file_txt} && -e $info{file_txt}) {
        say sprintf(" Document file     : %s", $info{file_txt});
    } else {
        say sprintf(" Document file     : %s", magenta("not found"));
    }

    # Signature file
    if ($info{file_sig} && -e $info{file_sig}) {
        say sprintf(" Signature file    : %s", $info{file_sig});
    } else {
        say sprintf(" Signature file    : %s", magenta("not found"));
    }

    # Source file
    if ($info{file_raw} && -e $info{file_raw}) {
        say sprintf(" Source file       : %s", $info{file_raw});
        say sprintf("   File size       : %s", fbytes(filesize($info{file_raw})));
        say sprintf("   File md5        : %s", md5sum($info{file_raw}));
        say sprintf("   File sha256     : %s", sha256sum($info{file_raw}));
    } else {
        say sprintf(" Source file       : %s", magenta("not found"));
    }

    # Verify
    my $status = 0;
    if ($info{file_sig} && -e $info{file_sig}) {
        if ($self->verify_file($info{file_sig}, $info{file_txt}, !$self->verbosemode)) {
            $status = 1;
            say green("Document is valid");
        } else {
            say red("Document is invalid");
        }
    } else {
        say yellow("No signature found");
    }

    return $status;
});

__PACKAGE__->register_handler(
    handler     => "show",
    description => "Search noRFC document by number or show list of documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments)
        || $self->cli_prompt('Please provide the noRFC number (e.g., NoRFC-R000000000):'));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;

    # Extract document from server
    my %info = $self->extract($docnum);
    return 0 unless %info;
    return $self->raise("Document not found") unless $info{id};

    # Found!
    say sprintf("Found document     : %s", $docnum);
    say sprintf(" Id                : %s", $info{id});
    say sprintf(" Category          : %s", encode(locale => $info{category} // ''));
    say sprintf(" Title             : %s", encode(locale => $info{title} // ''));
    say sprintf(" Published         : %s", $info{pub} ? fdate($info{pub}) : yellow("unknown"));
    say sprintf(" Updates           : %s", join(", ", @{($info{updates})})) if isnt_void($info{updates}) && is_array($info{updates});
    say sprintf(" Obsoletes         : %s", join(", ", @{($info{obsoletes})})) if isnt_void($info{obsoletes}) && is_array($info{obsoletes});
    say sprintf(" Author            : %s", encode(locale => $info{author} // ''));
    say sprintf(" Attributes        : %s", $info{attributes} // '');
    say sprintf(" Document file     : %s", $info{file_txt}) if $info{file_txt};
    say sprintf(" Signature file    : %s", $info{file_sig}) if $info{file_sig};
    if ($info{file_raw}) {
        say sprintf(" Source file       : %s", $info{file_raw});
        say sprintf("   File size       : %s", fbytes(filesize($info{file_raw})));
        say sprintf("   File md5        : %s", md5sum($info{file_raw}));
        say sprintf("   File sha256     : %s", sha256sum($info{file_raw}));
    }
    my %attr = attr2hash($info{attributes});

    # Verify
    if ($info{file_sig} && -e $info{file_sig}) {
        if ($self->verify_file($info{file_sig}, $info{file_txt}, !$self->verbosemode)) {
            say green("Document is valid");
        } else {
            say red("Document is invalid");
        }
    } else {
        #return $self->raise("Document is invalid");
        say yellow("No signature found");
    }

    # Pager
    my $pager = CTK::Util::syscfg("pager") || CTK::Util::which("less") || CTK::Util::which("more") || CTK::Util::which("cat");

    # Hexdumper
    my $hexdumper = CTK::Util::which("hexdump") || CTK::Util::which("xxd") || '';
    if ($attr{binary} && $hexdumper && -e $hexdumper) {
        unless ($self->option("force")) {
            say yellow("Source file is binary. Use --force option to show hexdump of the source file");
            return 1;
        }
        if ($pager && -e $pager) {
            my @cmd = ($hexdumper);
            push @cmd, "-C" if $hexdumper =~ /hexdump$/;
            push @cmd, ($info{file_raw}, '|', $pager);
            system(join(" ", @cmd));
        }
        return 1;
    }

    # Show text document
    if ($pager && -e $pager) {
        my $cmd = [$pager, $info{file_txt}];
        system(@$cmd);
        my $stt = $? >> 8;
        if ($stt) {
            say cyan("#", join(" ", @$cmd));
            say red("Exitval=%d", $stt);
        }
    } else {
        say yellow("Pager not found") if $self->verbosemode;
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "search",
    description => "Search noRFC documents by text and show list of found documents",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $offset = $self->option("offset") || 0;
    my $page = $self->option("page") || 0;
    my $limit = $self->option("limit") || MAX_ROWS;

    # Correct offset by page
    $page = (( $offset - ($offset % $limit)) / $limit) + 1 if !$page && $offset;

    # Get data from server
    $self->client->search(
            text    => decode(locale => $self->option("title") // shift(@arguments) // ''),
            category=> decode(locale => $self->option("category") // ''),
            page    => $page,
            limit   => $limit,
        ) or do {
            $self->raise($self->client->error);
            $self->debug($self->client->trace);
            return 0;
    };

    # Debug
    $self->debug(sprintf("Message         : %s", $self->client->res->json("/message") || ''));
    $self->debug(sprintf("Count (all)     : %s", $self->client->res->json("/count") || 0));
    $self->debug(sprintf("Pages           : %s", $self->client->res->json("/pages") || 0));
    $self->debug(sprintf("Limit           : %s", $self->client->res->json("/limit") || 0));
    $self->debug(sprintf("Page (current)  : %s", $self->client->res->json("/page") || 0));
    $self->debug(sprintf("Total (rows)    : %s", $self->client->res->json("/total") || 0));
    $self->debug(sprintf("Category        : %s", $self->client->res->json("/in/category") || 'none'));
    $self->debug(sprintf("Text            : %s", $self->client->res->json("/in/text") || 'none'));
    my $found = array($self->client->res->json("/found"));
    #print explain $found;

    # Generate table
    my $len = (SCREENWIDTH() - 64);
       $len = 44 if $len < 44;
    my $tbl = Text::SimpleTable->new(
            [21,    'DOCUMENT NUMBER'],
            [32,    'CATEGORY'],
            [$len,  'TITLE'],
        );
    my $total = 0;
    foreach my $row (@$found) {
        $total++;
        $tbl->row(
            uv2null(value($row => "num")),
            variant_stf(encode(locale => uv2null(value($row => "category"))), 30),
            variant_stf(encode(locale => uv2null(value($row => "title"))), $len-2),
        );
    }

    # Show table
    unless ($total) {
        say yellow("Your search string did not match any documents");
        return 1;
    }
    print $tbl->draw();
    my $tf = sprintf("row(s) %d of %d; page %d of %d",
        $self->client->res->json("/total") || 0,
        $self->client->res->json("/count") || 0,
        $self->client->res->json("/page") || 1,
        $self->client->res->json("/pages") || 1,
    );
    printf "%*s\n\n", (SCREENWIDTH() - 2), $tf;

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "del",
    description => "Delete noRFC document by number",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments)
        || $self->cli_prompt('Please provide the noRFC number (e.g., NoRFC-R000000000):'));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;

    # Remove document from server (by $docnum)
    $self->client->remove($docnum) or do {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    };
    # Remove document from server (by $number)
    if ($docnum ne $number) {
        $self->client->remove($number) or do {
            $self->raise($self->client->error);
            $self->debug($self->client->trace);
            return 0;
        };
    }
    unless ($self->client->res->json("/status")) {
        say yellow("Skipped. The document %s not found", $docnum);
        return 1;
    }

    # Remove relations
    $self->client->reldel($docnum) or do {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    };

    # Show result
    say cyan("The document %s has been successfully removed", $docnum);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "relation",
    description => "Manipulations with relations",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $cmd = lc(shift(@arguments) // 'none');
    return $self->raise("Subcommand incorrect") unless grep {$cmd eq $_} qw/add get del show/;

    # Get number and create norfcnum object
    my $number = trim(shift(@arguments)
        || $self->cli_prompt('Please provide the noRFC number (e.g., NoRFC-R000000000):'));
    my $norfcnum = $self->is_valid($number) or return 0;
    my $docnum = $norfcnum->as_string;

    # Get relations from console options
    my %relations = $self->relations;

    # Subcommands
    my $status;
    if ($cmd eq 'add') {
        unless (keys %relations) {
            say yellow("No `Updates` and `Obsoletes` specified");
            return 1;
        }
        while (my ($ext, $name) = each %relations) {
            $status = $self->client->reladd($docnum, $ext, $name);
            last unless $status;
        }
        # Show result
        if ($self->client->res->json("/status")) {
            say cyan("The relation(s) for document %s has been successfully added", $docnum);
        }

        #say $self->client->res->body;
    } elsif ($cmd eq 'get' or $cmd eq 'show') {
        my @found = ();

        # Each rel
        foreach my $ext (keys %relations) {
            $status = $self->client->relget($docnum, $ext);
            last unless $status;
            my $f = array($self->client->res->json("/relations"));
            push @found, @$f;
        }

        # All rels
        unless (keys %relations) {
            $status = $self->client->relget($docnum);
            if ($status) {
                my $f = array($self->client->res->json("/relations"));
                push @found, @$f;
            }
        }

        # Show table
        #say explain(\@found);
        if (@found) {
            my $tbl = Text::SimpleTable->new(
                [5,     'ID'],
                [21,    'DOCUMENT NUMBER'],
                [21,    'EXTERNAL NUMBER'],
                [32,    'NAME'],
            );
            foreach my $r (@found) {
                $tbl->row(uv2null(value($r, "id")), uv2null(value($r, "number")),
                    uv2null(value($r, "external")), uv2null(value($r, "name")));
            }
            print $tbl->draw();
        } else {
            say yellow("Relations not found");
        }
    } elsif ($cmd eq 'del') {
        # Each rel
        foreach my $ext (keys %relations) {
            $status = $self->client->reldel($docnum, $ext);
            last unless $status;
        }
        # All rels
        unless (keys %relations) {
            $status = $self->client->reldel($docnum);
        }

        # Show result
        if ($self->client->res->json("/status")) {
            say cyan("The relation(s) for document %s has been successfully removed", $docnum);
        }
    }

    # Check client status
    unless ($status) {
        $self->raise($self->client->error);
        $self->debug($self->client->trace);
        return 0;
    }

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "clean",
    description => "Cleaning temporary directories",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Remofe files from temp dir
    $self->debug(sprintf("Remove all files from temp dir: %s", $self->tempdir));
    find({  follow      => 1,
            follow_skip => 2,
            no_chdir    => 0,
            wanted      => sub
    {
        my $fn = $_;
        my $dir = $File::Find::dir;
        if ($fn =~ /^(Draft|NoRFC)/) {
            return 0 unless -f $fn;
            #say cyan("Dir=%s; File=%s", $dir, $fn);
            unlink $fn or warn "Could not unlink $fn: $!";
        }
    }}, $self->tempdir);

    # Remove dbdir
    my $dbdir = File::Spec->catdir($self->datadir, DBDIR);
    say cyan("Database directory: %s", $dbdir);
    if ($self->option("force") || $self->cli_prompt("Are you sure you want to remove all files from database directory?:", 'no') !~ /^\s*y/i) {
        return $self->raise("Operation aborted");
    }
    $self->debug(sprintf("Remove all files from DB dir: %s", $dbdir));
    rmtree($dbdir, {keep_root => 1});

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "truncate",
    description => "Delete all records from database",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store
    my $store = $self->store;
    return $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;

    # Request
    if ($self->cli_prompt(sprintf("Delete all records from database (%s)?:", $store->dsn), 'no') !~ /^\s*y/i) {
        return $self->raise("Operation aborted");
    }

    # Truncate
    $store->truncate() or return $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn));

    # Show result
    say cyan("Database was cleared");

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "export",
    description => "Export all noRFC documents from database",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;
    my $status = 1; # Ok by default

    # DB store
    my $store = $self->store;
    return $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;

    # Get count
    my $count = $store->count();
    return $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;

    # Get max_rows (limit)
    my $limit = $self->option("limit") || MAX_ROWS;
       $limit = MAX_ROWS unless is_int($limit);

    # Get pages number
    my $pages = (($count - ($count % $limit)) / $limit) + (($count % $limit) ? 1 : 0);

    # Load history data
    my $yfile = File::Spec->catfile($self->datadir, YFILE);
    my $history = $self->yaml_load($yfile) || {};
    return $self->raise($self->error) if $self->error;

    # Make index data
    my $dkey = dtf(DATE_FORMAT);
    my %index_data = (
        dkey    => $dkey, # Date from history
        dcnt    => $history->{$dkey} || 0, # Day's counter from history
        rows    => $count, # Number of documents in database
        max_rows=> $limit, # Max rows perl page for importing
        pages   => $pages, # Pages number
        total   => 0,      # Real number of documents that was exported
    );
    #print explain(\%index_data);

    # Snapshot directory
    my $snapshot_dir = File::Spec->catdir($self->tempdir, SNAPSHOTDIR, $dkey);
    preparedir($snapshot_dir) or return $self->raise(sprintf("Can't prepare directory %s", $snapshot_dir));

    # External loop (pages)
    my $total = 0;
    my $bar = NoRFC::Progress->new();
    $bar->start(sprintf("Exporting %d documents (%d pages)... ", $count, $pages), $total+1, $count);
    for (my $offset = 0; $offset < $count; $offset += $limit) {
        my $rows = ($offset + $limit > $count) ? $count % $limit : $limit;

        # Get data from database
        my @all = $store->getall($offset, $rows);
        unless ($store->status) {
            $bar->finish(red("Fail"));
            $status = $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
            last;
        }

        # Internal loop (documents)
        foreach my $row (@all) {
            next unless is_array($row); # Skip all incorrect records
            my $norfcnum = NoRFC::Number->new($row->[1]);
            next unless $norfcnum->is_valid; # Skip all invalid numbers
            my $docnum = $norfcnum->as_string;
            $bar->update(sprintf("[%s] ", $docnum), ++$total);
            my $subdir = File::Spec->catdir($norfcnum->year, sprintf("%d%03d", $norfcnum->year, $norfcnum->yday)); # <Y3>/<Y3><YDAY>
            my $target_dir = File::Spec->catdir($snapshot_dir, $subdir);
            preparedir($target_dir) or do {
                $bar->finish(red("Fail"));
                $status = $self->raise(sprintf("Can't prepare directory %s", $target_dir));
                last;
            };

            # Get relations
            my @relations = $store->relget($docnum);
            unless ($store->status) {
                $bar->finish(red("Fail"));
                $status = $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
                last;
            }

            # Save file
            my $ymlfile = File::Spec->catfile($target_dir, sprintf("%s.yml", $docnum));
            $self->yaml_save($ymlfile, {
                    id          => $row->[0],
                    num         => $row->[1],
                    pub         => $row->[2],
                    category    => $row->[3],
                    title       => $row->[4],
                    author      => $row->[5],
                    attributes  => $row->[6],
                    content     => $row->[7],
                    signature   => $row->[8],
                    relations   => [@relations],
                }) or do {
                    $bar->finish(red("Fail"));
                    $status = $self->raise($self->error);
                    last;
            };

            # Copy files
            my $source_dir = File::Spec->catdir($self->datadir, DBDIR, $subdir);
            if (-e $source_dir) {
                foreach my $ext (qw/raw raw.md5/) {
                    my $filename = sprintf("%s.%s", $docnum, $ext);
                    my $file_src = File::Spec->catfile($source_dir, $filename);
                    my $file_trg = File::Spec->catfile($target_dir, $filename);
                    next unless -e $file_src;
                    path($file_src)->copy_to($file_trg);
                }
            }
        }
        last unless $status;
    }
    $bar->finish(green("Done")) if $status;
    $self->debug(sprintf("Time elapsed: %s", $bar->elapsed));
    return 0 unless $status;

    # Save index data
    $index_data{total} = $total;
    my $index_file = File::Spec->catfile($snapshot_dir, INDEXFILE);
    $self->yaml_save($index_file, \%index_data) or return $self->raise($self->error);
    $self->debug(explain(\%index_data));

    # Compress directory
    print "Compress... ";
    print "\n" if $self->verbosemode;
    my $dirarch = path($snapshot_dir)->dirname->to_string;
    my $filearch = trim(shift(@arguments))
        || File::Spec->catfile(getcwd(), sprintf("norfc-%s.tgz", $dkey));
    my $exitval = _execmd(which(TARBIN) || TARBIN, "-c", "-z", $self->verbosemode ? ("-v") : (), "-C", $dirarch, "-f", $filearch, $dkey);
    if ($exitval) {
        say red("Fail");
        return $self->raise("Can't compress \"%s\" to \"%s\"", $snapshot_dir, $filearch);
    } else {
        if (-e $filearch) {
            say green("Done");
            rmtree($snapshot_dir, {keep_root => 0}) if -e $snapshot_dir;
        } else {
            say red("Fail");
            return $self->raise("Error while compress \"%s\" to \"%s\"", $snapshot_dir, $filearch);
        }
    }

    # Ok
    say "\n", cyan("All documents has been successfully exported to\n %s", $filearch);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "import",
    description => "Import all noRFC document to database",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # DB store
    my $store = $self->store;
    return $self->raise($store->error || sprintf("Store (%s): Unknown error", $store->dsn)) unless $store->status;

    # Resore directory
    my $restore_dir = File::Spec->catdir($self->tempdir, RESTOREDIR);
    preparedir($restore_dir, 0777) or return $self->raise(sprintf("Can't prepare directory %s", $restore_dir));

    # Tarball file
    my $tarball_file = shift(@arguments) // '';
    return $self->raise("No tarball file specified") unless length($tarball_file);
    return $self->raise("File not found") unless -e $tarball_file;

    # File of history data
    my $yfile = File::Spec->catfile($self->datadir, YFILE);

    # Extract directory
    printf "Extract %s to %s... ", $tarball_file, $restore_dir;
    print "\n" if $self->verbosemode;
    my $exitval = _execmd(which(TARBIN) || TARBIN, "-x", "-z", $self->verbosemode ? ("-v") : (), "-C", $restore_dir, "-f", $tarball_file);
    if ($exitval) {
        say red("Fail");
        return $self->raise("Can't extract \"%s\" to \"%s\"", $tarball_file, $restore_dir);
    } else {
        say green("Done");
    }

    # Get index file && restore_dir
    my $index_file;
    find({
        no_chdir    => 1,
        wanted      => sub
    {
        my $file = $File::Find::name; # $_;
        my $fn   = basename($file);
        if ($fn eq INDEXFILE) {
            #say cyan("Dir=%s; File=%s", $dir, $fn);
            $restore_dir = $File::Find::dir;
            $index_file = $file;
        }
    }}, $restore_dir);
    unless ($index_file && -e $index_file) {
        rmtree($restore_dir, {keep_root => 1}) if $restore_dir && -e $restore_dir;
        return $self->raise(sprintf("Index file %s not found", $index_file));
    }

    # Load index data
    my $index_data = $self->yaml_load($index_file) || {};
    return $self->raise($self->error) if $self->error;
    my $total = tv2int(value($index_data, "total") || value($index_data, "rows") || 0);
    unless ($total) {
        say yellow("No documents found");
        return 1;
    }

    # Main loop
    my $status = 1;
    my $i = 0;
    my $bar = NoRFC::Progress->new();
    $bar->start(sprintf("Importing %d documents... ", $total), 1, $total);
    find({
        no_chdir    => 1,
        wanted      => sub
    {
        return unless $status;
        my $dir = $File::Find::dir;
        my $file = $File::Find::name; # $_;
        my $fn   = basename($file);
        return if $fn eq INDEXFILE;
        return unless $fn =~ /\.yml$/;
        $i++;

        # Load info
        my $info = $self->yaml_load($file) || {};
        next if $self->error;

        # Get number
        my $docnum = value($info, "num");
        return unless $docnum;

        # Show bar
        $bar->update(sprintf("[%s] ", $docnum), $i);
        #say cyan("Dir=%s; File=%s; I=%d", $dir, $file, $i);

        # Number object
        my $norfcnum = NoRFC::Number->new($docnum);
        return unless $norfcnum->is_valid; # Skip all invalid numbers

        # Sub dir
        my $subdir = File::Spec->catdir($norfcnum->year, sprintf("%d%03d", $norfcnum->year, $norfcnum->yday)); # <Y3>/<Y3><YDAY>
        my $target_dir = File::Spec->catdir($self->datadir, DBDIR, $subdir);
        preparedir($target_dir) or do {
            $bar->finish(red("Fail"));
            $status = $self->raise(sprintf("Can't prepare directory %s", $target_dir));
            return;
        };

        # Copy files to db dir
        foreach my $ext (qw/raw raw.md5/) {
            my $filename = sprintf("%s.%s", $docnum, $ext);
            my $file_src = File::Spec->catfile($dir, $filename);
            my $file_trg = File::Spec->catfile($target_dir, $filename);
            next unless -e $file_src;
            path($file_src)->copy_to($file_trg);
        }

        # Check md5
        my %attr = attr2hash(value($info, "attributes"));
        my $file_raw = File::Spec->catfile($target_dir, sprintf("%s.raw", $docnum));
        if ($attr{external} && -e $file_raw) {
            my $md5_got = md5sum($file_raw);
            my $md5_exp = $attr{md5} || "unknown";
            unless ($md5_exp eq $md5_got) {
                $bar->finish(red("Bad file %s", $file_raw))->restart;
            }
        } elsif ($attr{external}) {
            $bar->finish(yellow("File %s not found", $file_raw))->restart;
        }

        # Add pure record
        $store->replace(%$info) or do {
            $bar->finish(red("Fail"));
            $status = $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
            return;
        };

        # Add relations
        my $relations = array($info, "relations");
        if (isnt_void($relations)) {

            # Remove all relations
            $store->reldel($docnum) or do {
                $bar->finish(red("Fail"));
                $status = $self->raise( $store->error || sprintf("Store (%s): Unknown error", $store->dsn) );
                return;
            };

            # Add relations
            foreach my $rel (@$relations) {
                next unless is_hash($rel);
                next if $store->reladd(
                    number  => $docnum,
                    external=> value($rel, "external"),
                    name    => value($rel, "name"),
                );
                $bar->finish(magenta("%s", $store->error || sprintf("Store (%s): Unknown error", $store->dsn)))->restart;
                $status = 0;
            }
        }
    }}, $restore_dir);
    $bar->finish(green("Done")) if $status;
    $self->debug(sprintf("Time elapsed: %s", $bar->elapsed));
    return 0 unless $status;

    # Save history data
    my %history = (
            value($index_data, "dkey") || dtf(DATE_FORMAT),
            tv2int(value($index_data, "dcnt") || 0)
        );
    $self->yaml_save($yfile, \%history) or return $self->raise($self->error);

    # Remove working directory
    rmtree($restore_dir, {keep_root => 0}) if $restore_dir && -e $restore_dir;

    # Ok
    say "\n", cyan("All documents has been successfully imported to\n %s", $store->dsn);

    return 1;
});

__PACKAGE__->register_handler(
    handler     => "server",
    description => "HTTP server",
    code => sub {
### CODE:
    my ($self, $meta, @arguments) = @_;

    # Destroy the store instance
    undef $self->{store}; # $self->{store} = NoRFC::Store->new(%store_args);

    # Get permisions by names
    my $uid = $>; # Effect. UID
    my $gid = $); # Effect. GID
    if (IS_ROOT) {
        $uid = getpwnam(USERNAME) || return $self->raise("getpwnam failed - $!");
        $gid = getgrnam(GROUPNAME) || return $self->raise("getgrnam failed - $!\n");
    }

    # Dash k
    my $dash_k = shift(@arguments) || "status";
    unless (grep {$_ eq $dash_k} qw/start status stop restart reload/) {
        return $self->raise("Incorrect LSB command! Please use start, status, stop, restart or reload");
    }

    # Set permisions (GID and UID) for work directory if not specified the DataDir directory in options
    unless ($self->option("datadir")) {
        chown($uid, $gid, $self->datadir) if IS_ROOT && File::stat::stat($self->datadir)->uid != $uid;
    }

    # Prepare tempdir
    $self->tempdir(File::Spec->catdir($self->tempdir, "server"));
    unless (preparedir( $self->tempdir, 0777 )) {
        return $self->raise(sprintf("Can't prepare temp directory: %s", $self->tempdir));
    }
    chown($uid, $gid, $self->tempdir) if IS_ROOT && File::stat::stat($self->tempdir)->uid != $uid;

    # Prepare log directory
    if (IS_ROOT) {
        my $logdir = $self->logdir;
        unless (preparedir( $logdir )) {
            return $self->raise(sprintf("Can't prepare log directory: %s", $logdir));
        }
        # Set permisions (GID and UID) for log directory
        chown($uid, $gid, $logdir) if File::stat::stat($logdir)->uid != $uid;
    } else {
        $self->logfile(File::Spec->catfile($self->tempdir(), sprintf("%s.log", PROJECTNAMEL)));
    }

    # Prepare pid directory and file
    my $piddir = IS_ROOT
        ? File::Spec->catdir( rundir(), PROJECTNAMEL)
        : $self->tempdir();
    my $pidfile = File::Spec->catfile($piddir, sprintf("%s.pid", PROJECTNAMEL));
    unless (preparedir($piddir)) {
        return $self->raise(sprintf("Can't prepare pid directory: %s", $piddir));
    }
    # Set permisions (GID and UID) for pid directory
    chown($uid, $gid, $piddir) if IS_ROOT && File::stat::stat($piddir)->uid != $uid;

    # Hypnotoad variables
    my $upgrade = 0;
    my $reload = 0;
    my $upgrade_timeout = UPGRADE_TIMEOUT;
    $ENV{MOJO_MODE} ||= $self->debugmode ? 'development' : 'production'; # This is a production server
    $ENV{MOJO_LOG_LEVEL} ||= $self->debugmode ? 'debug' : 'info';

    # Mojolicious Application
    my $app = NoRFC::Server->new(ctk => $self);
       $app->attr(ctk => sub { $self }); # has ctk => sub { CTKx->instance->ctk };
    my $prefork = Mojo::Server::Prefork->new( app => $app ); # app => $self
       $prefork->pid_file($pidfile);

    # Hypnotoad Pre-fork settings
    $prefork->max_clients(tv2int(value($self->conf("clients")))) if defined $self->conf("clients");
    $prefork->max_requests(tv2int(value($self->conf("requests")))) if defined $self->conf("requests");
    $prefork->accepts(tv2int(value($self->conf("accepts")))) if defined $self->conf("accepts");
    $prefork->spare(tv2int(value($self->conf("spare")))) if defined $self->conf("spare");
    $prefork->workers(tv2int(value($self->conf("workers")))) if defined $self->conf("workers");

    # Make Listen
    my $cfg_listen = value($self->conf("listen"));
    my $tls_on = isTrueFlag(value($self->conf("tls")));
    my $listen = $tls_on ? "https://" : "http://";
    if ($cfg_listen) {
        $listen .= $cfg_listen;
    } else {
        $listen .= sprintf("%s:%d",
                value($self->conf("listenaddr")) || SERVER_LISTEN_ADDR,
                tv2int16(value($self->conf("listenport"))) || SERVER_LISTEN_PORT,
            );
    }
    my $_resolve_cf = sub {
        my $f = shift;
        return $f if File::Spec->file_name_is_absolute($f);
        return File::Spec->catfile($self->root, $f);
    };
    if ($tls_on) {
        my @p = ();
        foreach my $k (qw/ciphers version/) {
            my $v = value($self->conf("tls_$k")) // '';
            next unless length $v;
            push @p, sprintf("%s=%s", $k, $v);
        }
        foreach my $k (qw/ca cert key/) {
            my $v = value($self->conf("tls_$k")) // '';
            next unless length $v;
            push @p, sprintf("%s=%s", $k, $_resolve_cf->($v));
        }
        push @p, sprintf("%s=%s", "verify", value($self->conf("tls_verify")) || '0x00')
            if value($self->conf("tls_verify"));
        $listen .= sprintf("?%s", join('&', @p));
    }
    $prefork->listen([$listen]);

    # Working with Dash k
    if ($dash_k eq 'start') {
        if (my $pid = $prefork->check_pid()) {
            say "Already running $pid";
            return 1;
        }
    } elsif ($dash_k eq 'stop') {
        if (my $pid = $prefork->check_pid()) {
            kill 'QUIT', $pid;
            say "Stopping $pid";
        } else {
            say "Not running";
        }
        return 1;
    } elsif ($dash_k eq 'restart') {
        if (my $pid = $prefork->check_pid()) {
            $upgrade ||= steady_time;
            kill 'QUIT', $pid;
            my $up = $upgrade_timeout;
            while (kill 0, $pid) {
                $up--;
                sleep 1;
            }
            die("Can't stop $pid") if $up <= 0;
            say "Stopping $pid";
            $upgrade = 0;
        }
    } elsif ($dash_k eq 'reload') {
        my $pid = $prefork->check_pid();
        if ($pid) {
            # Start hot deployment
            kill 'USR2', $pid;
            say "Reloading $pid";
            return 1;
        }
        say "Not running";
    } else { # status
        if (my $pid = $prefork->check_pid()) {
            say "Running $pid";
        } else {
            say "Not running";
        }
        return 1;
    }

    #
    # LSB start
    #

    # Listen USR2 (reload)
    local $SIG{USR2} = sub { $upgrade ||= steady_time };

    # Set hooks
    #$prefork->on(spawn => sub () { # Spawn (start worker)
    #    my $self = shift; # Prefork object
    #    my $pid = shift;
    #    #say "Spawn (start) $pid";
    #    $self->app->log->debug("Spawn (start) $pid");
    #});
    $prefork->on(wait => sub { # Manage (every 1 sec)
        my $self = shift; # Prefork object

        # Upgrade
        if ($upgrade) {
            #$self->app->log->debug(">>> " . $self->healthy() || '?');
            unless ($reload) {
                $reload = 1; # Off next reloading
                if ($self->app->reload()) {
                    $reload = 0;
                    $upgrade = 0;
                }
            }

            # Timeout
            if (($upgrade + $upgrade_timeout) <= steady_time()) {
                kill 'KILL', $$;
                $upgrade = 0;
            }
        }
    });
    #$prefork->on(reap => sub { # Cleanup (Emitted when a child process exited)
    #    my $self = shift; # Prefork object
    #    my $pid = shift;
    #    #say "Reap (cleanup) $pid";
    #    $self->app->log->debug("Reap (cleanup) $pid");
    #});
    $prefork->on(finish => sub { # Finish
        my $self = shift; # Prefork object
        my $graceful = shift;
        $self->app->log->debug($graceful ? 'Graceful server shutdown' : 'Server shutdown');
    });

    # Set GID and UID
    if (IS_ROOT) {
        if (defined($gid)) {
            POSIX::setgid($gid) || return $self->raise("setgid $gid failed - $!");
            $) = "$gid $gid"; # this calls setgroups
            return $self->raise("detected strange gid") if !($( eq "$gid $gid" && $) eq "$gid $gid"); # just to be sure
        }
        if (defined($uid)) {
            POSIX::setuid($uid) || return $self->raise("setuid $uid failed - $!");
            return $self->raise("detected strange uid") if !($< == $uid && $> == $uid); # just to be sure
        }
    }

    # Daemonize
    $prefork->daemonize() unless $self->debugmode();

    # Running
    say "Running";
    $prefork->run();

    #print explain($prefork);

    return 1;
});

sub again {
    my $self = shift;
       $self->SUPER::again(); # CTK::App again first!!

    # Reset datadir from config file
    if (my $newdatadir = $self->config("datadir")) {
        $self->datadir($newdatadir);
    }

    # Prepare DataDir
    preparedir($self->datadir()) or do {
        $self->raise("Can't prepare directory %s", $self->datadir());
    };

    # Store
    my $store_conf = $self->config("store") || {};
       $store_conf = {} unless is_hash($store_conf);
    my %store_args = %$store_conf;
    $store_args{file} = File::Spec->catfile($self->datadir, DBFILE)
        unless ($store_args{file} || $store_args{dsn});
    $store_args{compress} = $self->option("compress") || lvalue($self->config("compress")) || 0;
    $self->{store_args} = {%store_args};
    $self->{store} = NoRFC::Store->new(%store_args);
    #$self->debug(explain($store));

    # Client instance
    $self->{client} = NoRFC::Client->new(
        url                 => lvalue($self->conf("serverurl")) || SERVER_URL,
        username            => lvalue($self->conf("username") // ""), # For basic authorization
        password            => lvalue($self->conf("password") // ""), # For basic authorization
        ask_credentials     => 1,
        insecure            => $self->option("insecure") || lvalue($self->conf("insecure")),
        max_redirects       => lvalue($self->conf("maxredirects")),
        connect_timeout     => lvalue($self->conf("connecttimeout")),
        inactivity_timeout  => lvalue($self->conf("inactivitytimeout")),
        request_timeout     => $self->conf("requesttimeout"),
    );

    # Fingerprint
    $self->{fingerprint} = $self->option("fingerprint") || lvalue($self->conf("fingerprint")) || "";

    # Create GnuPG instance
    $self->{gnupg} = GnuPG::Interface->new();

    # Categories
    my $categories = lvalue($self->conf("categories")) // "";
    my @extra_categories = parsewords($categories); # qw/foo bar baz/
    @extra_categories = @{(EXTRA_CATEGORIES_MAP)} unless @extra_categories;
    $self->{categories} = [@{(NORFC_CATEGORIES_MAP)}];
    $self->{xcategories} = [@extra_categories];

    # Credentials
    my $credentials = array($self->conf("credential"));
    $self->{credentials} = $credentials;
    #print explain($credentials);

    return $self; # CTK requires!
}
sub raise {
    my $self = shift;
    say STDERR red(@_);
    return 0;
}
sub store {
    my $self = shift;
    return $self->{store}
}
sub client {
    my $self = shift;
    return $self->{client}
}
sub categories {
    my $self = shift;
    my $t = shift // 0;
    if ($t == 1) { # common
        return (@{($self->{categories})});
    } elsif ($t == 2) { # extra
        return (@{($self->{xcategories})});
    }
    # both
    return (@{($self->{categories})}, @{($self->{xcategories})});
}
sub relations {
    my $self = shift;

    # Get updates and obsoletes
    my $updates = array($self->option("updates"));
    my $obsoletes = array($self->option("obsoletes"));
    my %relations = ();
    foreach my $num (@$updates) {
        my $obj = NoRFC::Number->new($num);
        unless ($obj->is_valid) {
            say yellow("The %s number for list of updates is incorrect and has been skipped", $num);
            next;
        }
        my $doc = $obj->as_string;
        $relations{$doc} = "Updates";
    }
    foreach my $num (@$obsoletes) {
        my $obj = NoRFC::Number->new($num);
        unless ($obj->is_valid) {
            say yellow("The %s number for list of obsoletes is incorrect and has been skipped", $num);
            next;
        }
        my $doc = $obj->as_string;
        $relations{$doc} = "Obsoletes";
    }

    return (%relations);
}
sub gnupg {
    my $self = shift;
    return $self->{gnupg}
}
sub fingerprint {
    my $self = shift;
    my $value = shift;
    return uv2null($self->{fingerprint}) unless defined($value);
    $self->{fingerprint} = $value;
    return $value;
}
sub is_valid {
    my $self = shift; # First argument
    my $number = shift; # Last argument
    my $_verbose = shift;
    my $verbose = $_verbose ? 0 : 1; # 1 - silent; 0 - verbose (default)

    # Number
    unless ($number) {
        return $self->raise("No noRFC document number") if $verbose;
        return 0;
    }

    # Object
    my $norfcnum = NoRFC::Number->new($number);

    # YEAR
    unless ($norfcnum->year) {
        return $self->raise("The YEAR field of the noRFC document number is incorrect") if $verbose;
        return 0;
    }

    # YDAY
    unless ($norfcnum->yday) {
        return $self->raise("The YDAY field of the noRFC document number is incorrect") if $verbose;
        return 0;
    }

    # NN
    unless ($norfcnum->nn) {
        return $self->raise("The NN field of the noRFC document number is incorrect") if $verbose;
        return 0;
    }

    # M11R valid
    unless ($norfcnum->is_valid) {
        return $self->raise("The noRFC document number is NOT valid") if $verbose;
        return 0;
    }

    return $norfcnum; # OK
}
sub slurp_io {
    my $self = shift;
    my $io = shift;
    unless ($io) {
        return "" if -t STDIN;
        $io = IO::Handle->new();
        return "" unless $io->fdopen(fileno(STDIN), "r");
    }

    # Slurp
    my $contents = join '', $io->getlines;
    $io->close;
    return $contents;
}
sub private_keys_info {
    my $self = shift;
    my %users = (); # fingerprint => {name, validity, order}
    my $order = 0;
    my @secret_keys = $self->gnupg->get_secret_keys();
    foreach my $key (@secret_keys) {
        my $fp = $key->fingerprint->as_hex_string;
        my $validity = sprintf("%s - %s", $key->creation_date_string || "Unknown", $key->expiration_date_string || "Never");
        my @user_ids = $key->user_ids;
        my @user_names = ();
        foreach my $user (@user_ids) {
            push @user_names, $user->as_string;
        }
        my $u_users = join("\n", @user_names);
        $users{$fp} = {
            name => decode(locale => $u_users),
            validity => $validity,
            order => $order++,
        };
    }
    return (%users);
}
sub extract {
    my $self = shift;
    my $num = shift;
    my $dir = shift;
    my %res = ();
    $self->error("");

    # Check number
    unless ($num) {
        $self->error("No document number");
        return ();
    }

    # Norfcnum obj
    my $norfcnum = NoRFC::Number->new($num);
    my $docnum = $norfcnum->as_string;

    # Get info from server
    my $cli_status = $self->client->info($docnum);
    unless ($cli_status) {
        $self->error($self->client->error);
        $self->debug($self->client->trace);
        return ();
    }
    unless ($self->client->res->json("/id")) {
        $self->error("Document not found");
        return ();
    }
    $res{id} = $self->client->res->json("/id");
    $res{num} = $self->client->res->json("/num") || $docnum;
    $res{category} = $self->client->res->json("/category") // '';
    $res{title} = $self->client->res->json("/title") // '';
    $res{pub} = $self->client->res->json("/pub");
    $res{author} = $self->client->res->json("/author") // '';
    $res{signature} = $self->client->res->json("/signature") // '';
    $res{attributes} = $self->client->res->json("/attributes") // '';
    my %attr = attr2hash($res{attributes});

    # Get relations
    $res{updates} = array($self->client->res->json("/updates"));
    $res{obsoletes} = array($self->client->res->json("/obsoletes"));

    # Directory
    my $subdir = File::Spec->catdir($norfcnum->year, sprintf("%d%03d", $norfcnum->year, $norfcnum->yday)); # <Y3>/<Y3><YDAY>
    $dir ||= File::Spec->catdir($self->tempdir, DBDIR, $subdir);
    unless (preparedir($dir)) {
        $self->error(sprintf("Can't prepare directory %s", $dir));
        return ();
    }
    $res{dir} = $dir;

    # Filenames
    my $file_raw = File::Spec->catfile($dir, sprintf("%s.raw", $docnum));
    my $file_dec = File::Spec->catfile($dir, sprintf("%s.dec", $docnum));
    my $file_txt = File::Spec->catfile($dir, sprintf("%s.txt", $docnum));
    my $file_sig = File::Spec->catfile($dir, sprintf("%s.txt.sig", $docnum));
    $res{file_raw} = $file_raw;
    $res{file_dec} = $file_dec;
    $res{file_txt} = $file_txt;
    $res{file_sig} = $file_sig;

    # Download raw file from server
    $cli_status = $self->client->download(sprintf("%s.raw", $docnum) => $file_raw);
    unless ($cli_status) {
        $self->error($self->client->error);
        $self->debug($self->client->trace);
        return ();
    }

    # Save signature
    if (length($res{signature})) {
        unless (bsave($file_sig, $res{signature})) {
            $self->error(sprintf("Can't save file \"%s\"", $file_sig));
            return ();
        }
    }

    # Decription
    if ($attr{encrypted}) {
        $self->decrypt_file($file_raw, $file_dec);
        return () if $self->error;
    }

    # Create document
    my $doc = NoRFC::Document->new(
        columns     => $attr{columns},
        tabstop     => $attr{tabstop},
        number      => $docnum,
        category    => $res{category} || DEFAULT_CATEGORY(),
        published   => $res{pub},
        author      => $res{author} || DEFAULT_AUTHOR(),
        title       => $res{title} || DEFAULT_TITLE(),
        updates     => $res{updates},
        obsoletes   => $res{obsoletes},
        file        => $attr{encrypted} ? $file_dec : $file_raw,
    );
    #if ($attr{binary}) {
    unless ($doc->save($file_txt)) {
        $self->error(sprintf("Can't save file \"%s\"", $file_txt));
        return ();
    }

    # Return
    return %res;
}
sub sign {
    my $self = shift;
    my $fp = shift;
    my $document = shift // '';
    $self->error("");
    unless (_fingerprint_test($fp)) {
        $self->error("Fingerprint incorrect");
        return;
    }
    unless (length($document)) {
        $self->error("No document");
        return;
    }

    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );
    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );
    $self->gnupg->options->hash_init(
        armor   => 1,
    );
    $self->gnupg->options->extra_args([
            "--detach-sign",
            sprintf("--local-user=%s", $fp),
        ]);

    # This sets up the communication
    my $pid = $self->gnupg->sign( handles => $handles );

    # This passes in the plaintext
    #binmode($input, ':raw:utf8');
    $input->print(encode(locale => $document));
    $input->close; # this closes the communication channel, indicating we are done

    # Slurp oututs
    my $sign = $self->slurp_io($output);
    my $errs = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    # Check signature format
    if ($sign !~ /^\-\-\-\-\-BEGIN\sPGP.+\-\-\-\-\-END\sPGP/s) {
        $self->error($errs);
        return;
    }

    # Ok
    return $sign;
}
sub sign_file {
    my $self = shift;
    my $fp = shift; # fingerprint
    my $file_doc = shift;
    my $file_sig = shift;
    $self->error("");

    # Check fingerprint
    unless (_fingerprint_test($fp)) {
        $self->error("Fingerprint incorrect");
        return;
    }

    # Fake handlers
    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );
    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );

    # This sets up the communication
    my $pid = $self->gnupg->wrap_call(
        handles     => $handles,
        commands    => [
            "-a", # armor
            "-b", # detach-sign
            "-u" => $fp, # Local user
            "-o" => $file_sig, # Signature
            $file_doc, # Source file
        ]
    );

    # Slurp oututs
    my $out = $self->slurp_io($output);
    my $err = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    # Set error
    $self->error($err) if $err;

    # Debug output
    $self->debug($out);

    # Return output
    return $out // "";
}
sub verify_file {
    my $self = shift;
    my $file_sig = shift;
    my $file_doc = shift;
    my $_verbose = shift;
    my $verbose = $_verbose ? 0 : 1; # 1 - silent; 0 - vrbose (default)

    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );
    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );

    # This sets up the communication
    my $pid = $self->gnupg->wrap_call(
        handles => $handles,
        commands => [
            "--verify",
            $file_sig,
            $file_doc
        ]);

    # Slurp oututs
    my $out = $self->slurp_io($output);
    say yellow($out) if $verbose && $out;
    my $err = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    my $status = 0; # Error
    if ($err =~ /Good/) {
        say yellow($err) if $verbose;
        $status = 1;
    } elsif ($err) {
        say red($err) if $verbose;
    }
    if ($out =~ /Good/) {
        $status = 1;
    }

    return $status;
}
sub encrypt_file {
    my $self = shift;
    my $fp = shift; # fingerprint
    my $file_dec = shift;
    my $file_enc = shift;
    $self->error("");

    # Check fingerprint
    unless (_fingerprint_test($fp)) {
        $self->error("Fingerprint incorrect");
        return;
    }

    # Fake handlers
    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );

    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );

    # This sets up the communication
    my $pid = $self->gnupg->wrap_call(
        handles     => $handles,
        commands    => [
            "-e", "-q", # "-a",
            "-r" => $fp,
            "-o" => $file_enc,
            $file_dec,
        ]
    );

    # Slurp oututs
    my $out = $self->slurp_io($output);
    my $err = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    # Set error
    $self->error($err) if $err;

    # Debug output
    $self->debug($out);

    # Return output
    return $out // "";
}
sub decrypt_file {
    my $self = shift;
    my $file_enc = shift;
    my $file_dec = shift;
    $self->error("");

    # Fake handlers
    my ($input, $output, $error) = (
            IO::Handle->new(),
            IO::Handle->new(),
            IO::Handle->new(),
        );

    my $handles = GnuPG::Handles->new(
            stdin    => $input,
            stdout   => $output,
            stderr   => $error,
        );

    # This sets up the communication
    my $pid = $self->gnupg->wrap_call(
        handles     => $handles,
        commands    => [
            "-d", "-q",
            "-o" => $file_dec,
            $file_enc,
        ]
    );

    # Slurp oututs
    my $out = $self->slurp_io($output);
    my $err = $self->slurp_io($error);

    # Clean up the finished GnuPG process
    waitpid $pid, 0;

    # Set error
    $self->error($err) if $err;

    # Debug output
    $self->debug($out);

    # Return output
    return $out // "";
}
sub yaml_load {
    my $self = shift;
    my $yfile = shift;
    my $err = "";
    $self->error("");

    my $data;
    if (-e $yfile and -s _) {
        try {
            $data = LoadFile($yfile);
        } catch {
            $err = sprintf("Can't load YAML file %s: %s", $yfile, $_ // "unknown error");
        };
        if ($err) {
            $self->error($err);
            return undef;
        }
    }
    return $data;
}
sub yaml_save {
    my $self = shift;
    my $yfile = shift;
    my $data = shift;
    my $err = "";
    $self->error("");

    if ($yfile) {
        try {
            DumpFile($yfile, $data);
        } catch {
            $err = sprintf("Can't save YAML file \"%s\": %s", $yfile, $_ // "unknown error");
        };
        if ($err) {
            $self->error($err);
            return 0;
        }
    }
    return 1;
}

# Getters
sub getType {
    my $self = shift;

    # Document type
    my $type = "";

    # Try get from options first
    if (my $t = $self->option("type")) {
        if ($t =~ /draft/i) {
            $type = "Draft";
        } else {
            $type = "NoRFC";
        }
    }

    # Ask
    unless ($type) {
        say "Available document types:";
        say "  N - NoRFC document";
        say "  D - Draft of noRFC document (default)";
        my $t = trim($self->cli_prompt('Please provide the document type (N/D):','D'));
        if ($t =~ /^d/i) {
            $type = "Draft";
        } elsif ($t =~ /^n/i) {
            $type = "NoRFC";
        } else {
            $type = "NoRFC";
        }
    }

    # Debug
    #$self->debug(sprintf("Document type: %s", $type));

    return $type;
}
sub getStatus {
    my $self = shift;

    # Status
    my $let = "";

    # Try get from options first
    $let .= "R" if $self->option("regular");
    $let .= "P" if $self->option("private");
    $let .= "X" if $self->option("extra");
    $let .= "Z" if $self->option("secret");

    # Ask
    unless ($let) {
        say "Available document statuses:";
        say "  R - Regular public document (default)";
        say "  P - Private document";
        say "  X - Extra protected document";
        say "  Z - Top secret document";
        $let = uc(trim($self->cli_prompt('Please type a char of the status (R/P/X/Z):','R')));
    }

    # Check
    if ($let !~ /^[RPXZ]$/) {
        return $self->raise("Incorrect the letter char selected! Please specify ONE flag -R, -P, -X or -Z only! Default: -R");
    }

    # Debug
    #$self->debug(sprintf("Status letter: %s", $let));

    return $let;
}
sub getCategory {
    my $self = shift;
    my $def = shift // DEFAULT_CATEGORY;

    # Category from options
    my $_category = $self->option("category");
    unless ($_category) {
        say "Available recommended document categories:";
        printf "  NoRFC cats: %s\n", join(", ", $self->categories(1));
        printf "  Extra cats: %s\n", join(", ", $self->categories(2));
        $_category = $self->cli_prompt('Please provide the category of the document:', $def);
    }
    my $category = decode(locale => $_category);
    if (length($category) > 64) {
        return $self->raise("The category value is too long! Maximum: 64 chars");
    }

    # Debug
    #$self->debug(sprintf("Category: %s", $_category));

    return $category;
}
sub getTitle {
    my $self = shift;
    my $def = shift // DEFAULT_TITLE;

    # Title from options
    my $_title = $self->option("title")
        || $self->cli_prompt('Please provide the title of the document:', $def);
    my $title = decode(locale => $_title);
    if (length($title) > 255) {
        return $self->raise("The title value is too long! Maximum: 255 chars");
    }

    # Debug
    #$self->debug(sprintf("Title: %s", $_title));

    return $title;
}
sub getAuthor {
    my $self = shift;

    # Default fingerprint (config, as default)
    my $fp_default = lvalue($self->conf("fingerprint"));
    if ($fp_default && !_fingerprint_test($fp_default)) {
        return $self->raise("The fingerprint is incorrect (see configuration file)");
    }
    #$self->debug(sprintf("Default fingerprint: %s", $fp_default)) if $fp_default;

    # Fingerprint (from command line)
    my $fingerprint = $self->option("fingerprint");
    if ($fingerprint && !_fingerprint_test($fingerprint)) {
        return $self->raise("The user's fingerprint is incorrect (see command line syntax)");
    }
    #$self->debug(sprintf("Fingerprint from command line: %s", $fingerprint)) if $fingerprint;

    # Get author
    my $author; # Author
    my %users = $self->private_keys_info(); # fingerprint => {name, validity, order}
    if ($fingerprint && exists($users{$fingerprint})) {
        # Catched!
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
        #$self->debug(sprintf("@ %s", encode(locale => $author)));
    } elsif (%users) {
        # Select
        say "Available secret-keys:\n";
        my $default_key;
        foreach my $k (sort {$users{$a}{order} <=> $users{$b}{order}} keys %users) {
            $fp_default ||= $k;
            $default_key = $k if $k eq $fp_default;
            printf("%s [%s]%s\n", yellow($k), $users{$k}{validity}, $k eq $fp_default ? " (default)" : "");
            printf(" @ %s\n\n", encode(locale => $users{$k}{name}));
        }
        $fingerprint = $self->cli_prompt('Please select the secret-key (fingerprint):', $default_key);
        if ($fingerprint) {
            if ($fingerprint =~ /^\s*n/i) {
                $fingerprint = "";
                say yellow("No signature specified. Skipped");
            } elsif (!(_fingerprint_test($fingerprint) && $users{$fingerprint})) {
                $self->error("Incorrect fingerprint");
                say yellow("Type \"no\" to skip signing");
                return 0;
            }
        }
        $author = $users{$fingerprint}{name} || DEFAULT_AUTHOR;
        #$self->debug(sprintf("@ %s", encode(locale => $author)));
    } else {
        $author = DEFAULT_AUTHOR;
    }

    # Set fingerprint
    if ($fingerprint) {
        $self->fingerprint($fingerprint);
        #$self->debug(sprintf("Fingerprint: %s", $fingerprint));
    }

    return $author;
}

# Private functions
sub _fingerprint_test {
    my $fp = shift || '';
    my $l = length($fp);
    return 0 unless $l == 40 or $l == 16; # Fingerprint or KeyID
    return 1 if $fp =~ /^[0-9a-f]+$/i;
    return 0;
}
sub _execmd {
    my @cmd = (@_);
    require IPC::Open3;
    my $pid = IPC::Open3::open3(my $chld_in, '>&STDOUT', '>&STDERR', @cmd);
    waitpid($pid, 0);
    my $child_exit_status = $? >> 8;
    return $child_exit_status;
}


1;

__END__
