package NoRFC::Client;
use strict;
use feature qw/say/;
use utf8;

=encoding utf-8

=head1 NAME

NoRFC::Client - Client for interaction with noRFC server

=head1 VIRSION

Version 1.00

=head1 SYNOPSIS

    use NoRFC::Client;

    my $clinet = NoRFC::Client->new(
        url => "https://localhost:8732",
        username            => "username", # optional
        password            => "password", # optional
        max_redirects       => 2, # Default: 10
        connect_timeout     => 3, # Default: 10 sec
        inactivity_timeout  => 5, # Default: 30 sec
        request_timeout     => 10, # Default: 5 min (300 sec)
    );
    my $status = $client->check();

    if ($status) {
        print STDOUT $client->res->body;
    } else {
        print STDERR $clinet->error;
    }

=head1 DESCRIPTION

Client for interaction with NoRFC server

=head2 new

    my $clinet = NoRFC::Client->new(
        url => "https://localhost:8732",
        username            => "username", # optional
        password            => "password", # optional
        max_redirects       => 2, # Default: 10
        connect_timeout     => 3, # Default: 10 sec
        inactivity_timeout  => 5, # Default: 30 sec
        request_timeout     => 10, # Default: 5 min (300 sec)
    );

Returns client

=over 8

=item B<ask_credentials>

Enables ask username and password from terminal

=item B<max_redirects>

Maximum number of redirects the user agent will follow before it fails. Default - 10

=item B<password>

Default password for basic authentication

=item B<*timeout>

Timeout for connections, requests and inactivity periods in seconds.

=item B<ua>

The Mojo UserAgent object

=item B<url>

Full URL of the WEB Server

=item B<username>

Default username for basic authentication

=back

=head1 METHODS

=head2 check

    my $status = $client->check;
    my $status = $client->check(URL);

Returns check-status of server. 0 - Error; 1 - Ok

=head2 cleanup

    $client->cleanup;

Cleanup all variable data in object and returns client object

=head2 code

    my $code = $clinet->code;

Returns HTTP code of the response

=head2 credentials

    my $userinfo = $client->credentials($MOJO_URL_OBJECT)

Gets credentials for User Agent

=head2 download

    my $status = $client->download("Draft-R2022358171.txt", "/tmp/Draft-R2022358171.txt");

Request for download file from server by file path.
The method returns status of operation: 0 - Error; 1 - Ok

=head2 error

    print $clinet->error;

Returns error string

=head2 gendocnum

    my $status = $client->gendocnum(
        type    => "NoRFC",
        status  => "P",
        year    => 2022,
        yday    => 123,
        nn      => 1,
    );

Generate the new document number

=head2 info

    my $status = $clinet->info();
    my $status = $clinet->info("Draft-R2022358171");

Request for getting server information or information about specified document

=head2 publish

    my $status = $clinet->publish(
        number      => $docnum,
        published   => time,
        category    => $category,
        title       => $title,
        author      => $author,
        fileraw     => $file_raw,
        filesig     => $file_sig,
        attributes  => hash2attr(encrypted => 1),
    );

Publication the noRFC document

=head2 reladd

    my $status = $client->reladd("NoRFC-R202300913", "NoRFC-R202300945", "Obsoletes");

Add new named relation for specified document number

=head2 reldel

    my $status = $client->reldel("NoRFC-R202300913");
    my $status = $client->reldel("NoRFC-R202300913", "NoRFC-R202300945");

Delete relation by numbers

=head2 relget

    my $status = $client->reladd("NoRFC-R202300913");
    my $status = $client->reladd("NoRFC-R202300913", "NoRFC-R202300945");

Get relation by numbers

=head2 remove

    my $status = $client->remove("NoRFC-R202300913");

Request for deleting the noRFC document from server.
The method returns status of operation: 0 - Error; 1 - Ok

=head2 req

    my $request = $clinet->req;

Returns Mojo::Message::Request object

=head2 request

    my $json = $clinet->request("METHOD", "PATH", ...ATTRIBUTES...);

Send request

=head2 res

    my $response = $clinet->res;

Returns Mojo::Message::Response object

=head2 search

    my $status = $client->search(
        text    => "Test string",
        category=> "Etcetera",
        page    => 1,
        limit   => 10,
    )

Search documents by text and cetegory

=head2 status

    my $status = $clinet->status;

Returns object status value. 0 - Error; 1 - Ok

=head2 trace

    my $trace = $client->trace;
    print $client->trace("New trace record");

Gets trace stack or pushes new trace record to trace stack

=head2 tx

    my $status = $clinet->tx($tx);

Works with Mojo::Transaction object, interface with it

=head2 ua

    my $ua = $clinet->ua;

Returns Mojo::UserAgent object

=head2 url

    my $url_object = $clinet->url;

Returns Mojo::URL object

=head1 SEE ALSO

L<Mojolicious>, L<Mojo::UserAgent>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION /;
$VERSION = '1.00';

use Mojo::UserAgent;
use Mojo::Asset::File;
use Mojo::File qw/path/;
use Mojo::URL;
use Mojo::Util qw/steady_time/;

use CTK::Util qw/dtf lf_normalize touch/;

use NoRFC::Util qw/md5sum parse_credentials fbytes fduration/;
use NoRFC::Const;

use constant {
        MAX_REDIRECTS       => 10,
        CONNECT_TIMEOUT     => 10,
        INACTIVITY_TIMEOUT  => 30,
        REQUEST_TIMEOUT     => 180,
        TRANSACTION_MASK    => "%s %s >>> %s %s [%s in %s%s]", # GET /auth >>> 200 OK [1.04 KB in 0.0242 seconds (43.1 KB/sec)]
        CONTENT_TYPE        => 'application/json',
        DEFAULT_URL         => 'http://localhost',
    };

sub new {
    my $class = shift;
    my %args  = @_;

    # General
    $args{url} ||= ""; # base url
    $args{prefix} = "";
    if ($args{url}) {
        $args{url} =~ s/\/+$//; $args{url} .= "/"; # Correct slash
        $args{url} = Mojo::URL->new($args{url});
        $args{prefix} = $args{url}->path;
    }
    $args{status} = 1; # 0 - error, 1 - ok
    $args{error} = ""; # string
    $args{code} = 0;   # integer
    $args{trace} = []; # trace pool
    $args{tx_time} = 0;
    $args{req} = undef;
    $args{res} = undef;
    $args{username} ||= "";
    $args{password} ||= "";
    $args{ask_credentials} ||= 0;
    if ($args{url} && $args{username}) { # Set userinfo
        $args{url}->userinfo(sprintf("%s:%s", $args{username}, $args{password}))
    } elsif ($args{url}) {
        ($args{username}, $args{password}) = (parse_credentials($args{url}->to_unsafe_string));
    }

    # User Agent
    my $ua = $args{ua};
    unless ($ua) {
        # Create the instance
        $ua = Mojo::UserAgent->new(
                max_redirects       => $args{max_redirects} || MAX_REDIRECTS,
                connect_timeout     => $args{connect_timeout} || CONNECT_TIMEOUT,
                inactivity_timeout  => $args{inactivity_timeout} || INACTIVITY_TIMEOUT,
                request_timeout     => $args{request_timeout} || REQUEST_TIMEOUT,
                insecure            => $args{insecure} || 0,
            );
        $args{ua} = $ua;
    }

    my $self = bless {%args}, $class;
    return $self;
}
sub error {
    my $self = shift;
    my $e = shift;
    $self->{error} = $e if defined $e;
    return $self->{error};
}
sub status {
    my $self = shift;
    my $s = shift;
    $self->{status} = $s if defined $s;
    return $self->{status};
}
sub code {
    my $self = shift;
    my $c = shift;
    $self->{code} = $c if defined $c;
    return $self->{code};
}
sub trace {
    my $self = shift;
    my $v = shift;
    if (defined($v)) {
        my $a = $self->{trace};
        push @$a, lf_normalize($v);
        return lf_normalize($v);
    }
    my $trace = $self->{trace} || [];
    return join("\n",@$trace);
}
sub cleanup {
    my $self = shift;
    $self->{status} = 1;
    $self->{error} = "";
    $self->{code} = 0;
    undef $self->{req};
    $self->{req} = undef;
    undef $self->{res};
    $self->{res} = undef;
    undef $self->{trace};
    $self->{trace} = [];
    return $self;
}
sub req {
    my $self = shift;
    return $self->{req};
}
sub res {
    my $self = shift;
    return $self->{res};
}
sub url {
    my $self = shift;
    return $self->{url};
}
sub ua {
    my $self = shift;
    return $self->{ua};
}
sub tx {
    my $self = shift;
    my $tx = shift;

    # Check Error
    my $err = $tx->error;
    unless (!$err || $err->{code}) {
        $self->error($err->{message});
        $self->status(0);
    }
    $self->code($tx->res->code || "000");
    $self->status($tx->res->is_success ? 1 : 0);
    $self->error($tx->res->json("/message") || $err->{message} || "Unknown error" )
        if $tx->res->is_error && !$self->error;

    # Tracing
    my $length = $tx->res->body_size || 0;
    my $rtime = $self->{tx_time} // 0;
    $self->trace(sprintf(TRANSACTION_MASK,
        $tx->req->method, $tx->req->url->to_abs, # Method & URL
        $self->code, $tx->res->message || $err->{message} || "Unknown error", # Line
        fbytes($length), # Length
        fduration($rtime), # Duration
        $rtime ? sprintf(" (%s/sec)", fbytes($length/$rtime)) : "",
      ));
    my $req_hdrs = $tx->req->headers->to_string;
    if ($req_hdrs) {
        $self->trace(join("\n", map {$_ = "> $_"} split(/\n/, $req_hdrs)));
        $self->trace(">");
    }
    my $res_hdrs = $tx->res->headers->to_string;
    if ($res_hdrs) {
        $self->trace(join("\n", map {$_ = "< $_"} split(/\n/, $res_hdrs)));
        $self->trace("<");
    }

    # Request And Response
    $self->{req} = $tx->req;
    $self->{res} = $tx->res;

    return $self->status;
}
sub request {
    my $self = shift;
    my $meth = shift;
    my $_url = shift;
    my @params = @_;
    $self->cleanup(); # Cleanup first

    # Set URL + credentials
    my $url = $_url
        ? Mojo::URL->new("$_url")
        : $self->{url}
            ? $self->url->clone
            : Mojo::URL->new(DEFAULT_URL);
    $url->userinfo($self->credentials($url)) if $_url;

    # Request
    my $start_time = steady_time() * 1;
    my $tx = $self->ua->build_tx($meth, $url, @params); # Create transaction (tx)
    my $status = $self->tx($self->ua->start($tx)); # Run it and validate!);
    $self->{tx_time} = sprintf("%.*f",4, steady_time()*1 - $start_time) * 1;

    # Auth required
    if (!$status && $self->{ask_credentials} && $self->code == 401) {
        $self->cleanup();
        $url->userinfo($self->credentials($url, 1));

        # Request
        $tx = $self->ua->build_tx($meth, $url, @params); # Create transaction (tx)
        $status = $self->tx($self->ua->start($tx)); # Run it and validate!);
        $self->{tx_time} = sprintf("%.*f",4, steady_time()*1 - $start_time) * 1;

        # Check
        if (!$status && $self->code == 401) {
            $self->{username} = "";
            $self->{password} = "";
        }
    }


    return $status;
}
sub credentials {
    my $self = shift;
    my $url = shift;
       $url ||= $self->{url};
    my $ask = shift(@_) ? 1 : 0;
    my ($username, $password);

    # return predefined credentials
    return sprintf("%s:%s", $self->{username}, $self->{password}) if $self->{username} && length($self->{password});

    # return if url contains credentials
    ($username, $password) = (parse_credentials($url->to_unsafe_string));
    return sprintf("%s:%s", $username, $password) if $username && lengt($password);

    # prompt if ask flag is true and is terminal
    if ($ask && -t STDIN) {
        my $realm = 'noRFC server';
        printf STDERR "Enter username for %s at %s: ", $realm, $url->host_port;
        $username = <STDIN>;
        chomp($username);
        if (length($username)) {
            print STDERR "Password: ";
            system("stty -echo");
            $password = <STDIN>;
            system("stty echo");
            print STDERR "\n"; # because we disabled echo
            chomp($password);
            $self->{username} = $username;
            $self->{password} = $password;
        } else {
            return "";
        }
        return sprintf("%s:%s", $username, $password);
    }

    return "";
}

## SERVER METHODS
sub check {
    my $self = shift;
    my $url = shift;
    if (!$url && $self->{url}) {
        $url = $self->url->clone;
        $url->path("info");
    }
    return $self->request(HEAD => $url);
}
sub info {
    my $self = shift;
    my $num = shift;
    my $url = $num
        ? $self->url->clone->path(sprintf("info/%s", $num))
        : $self->url->clone->path("info");
    return $self->request(GET => $url);
}
sub gendocnum {
    my $self = shift;
    my %args = @_;
    my $url = $self->url->clone->path("gendocnum");
    return $self->request(GET => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
        },
        form => { %args },
    );
}
sub publish {
    my $self = shift;
    my %args = @_;
    my $url = $self->url->clone->path($args{number} || "/");

    # File (raw) - optional
    my $file_raw = $args{fileraw};
    my $asset_raw = Mojo::Asset::File->new(path => $file_raw);

    # File (sig) - optional
    my $file_sig = $args{filesig};

    # Request
    return $self->request(PUT => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
            'Content-Type' => 'multipart/form-data',
        },
        form => {
            published   => $args{published},
            category    => $args{category},
            title       => $args{title},
            author      => $args{author},
            attributes  => $args{attributes},

            # Files
            size    => $asset_raw->size,
            md5     => md5sum($asset_raw->path),
            $file_raw
                ? (fileraw => {
                        file        => $asset_raw,
                        filename    => path($file_raw)->basename,
                        'Content-Type' => 'application/octet-stream',
                    })
                : (),
            $file_sig
                ? (filesig => {
                        file        => Mojo::Asset::File->new(path => $file_sig),
                        filename    => path($file_sig)->basename,
                        'Content-Type' => 'text/plain',
                    })
                : (),
        },
    );
}
sub download {
    my $self = shift;
    my $name = shift; # Draft-R2022358171.txt, Draft-R2022358171.raw
    my $file = shift; # Full file name to save
    my $filepath = path($file);
    my $filename = $filepath->basename;
    my $url = $self->url->clone->path($name);
    my $status = $self->request(GET => $url);
    return $status unless $status;
    $self->res->save_to($file);
    return 1 if $filepath->stat->size;
    $self->error("Can't download file");
    return $self->status(0);
}
sub search {
    my $self = shift;
    my %args = @_;
    my $url = $self->url->clone->path("search");
    return $self->request(GET => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
        },
        form => { %args },
    );
}
sub remove {
    my $self = shift;
    my $num = shift;
    my $url = $self->url->clone->path($num);
    return $self->request(DELETE => $url);
}
sub reladd {
    my $self = shift;
    my $num = shift;
    my $ext = shift;
    my $name = shift;
    my $url = $self->url->clone->path(sprintf("rel/%s", $num // ''));
    return $self->request(POST => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
        },
        form => {
            external => $ext // '',
            name     => $name // '',
        },
    );
}
sub relget {
    my $self = shift;
    my $num = shift;
    my $ext = shift;
    my $url = $self->url->clone->path(sprintf("rel/%s", $num // ''));
    return $self->request(GET => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
        },
        form => {
            external => $ext // '',
        },
    );
}
sub reldel {
    my $self = shift;
    my $num = shift;
    my $ext = shift;
    my $url = $self->url->clone->path(sprintf("rel/%s", $num // ''));
    return $self->request(DELETE => $url =>
        { # Headers
            'User-Agent' => sprintf("%s/%s", __PACKAGE__, $self->VERSION),
        },
        form => {
            external => $ext // '',
        },
    );
}

1;

__END__
