package NoRFC::Document;
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Document - The noRFC document manipulations

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use NoRFC::Document;

=head1 DESCRIPTION

This module provides the noRFC document interface

=head2 new

    my $norfcdoc = NoRFC::Document->new;

Creates new object instance

=head2 author

    my $author = $norfcdoc->author;
    my $author = $norfcdoc->author("Serz Minus");

Sets and returns the author of the noRFC document

=head2 as_string

    my $string = $norfcdoc->as_string;

Returns the noRFC document as string

=head2 body

    my $body = $norfcdoc->body;
    my $body = $norfcdoc->body("Content");

Sets and returns the body of the noRFC document

=head2 category

    my $category = $norfcdoc->category;
    my $category = $norfcdoc->category("Unknown");

Sets and returns the category of the noRFC document

=head2 file

    my $file = $norfcdoc->file;
    my $file = $norfcdoc->file("/path/to/file.txt");

Sets and returns the source file of the noRFC document

=head2 is_binary

    my $binary = $norfcdoc->is_binary;

Returns 1 if source of the noRFC document is binary file

=head2 number

    my $number = $norfcdoc->number;
    my $number = $norfcdoc->number("NoRFC-R2020162124");

Sets and returns the number of the noRFC document

=head2 obsoletes

    my $obsoletes = $norfcdoc->obsoletes;

Sets and returns the obsoletes of the noRFC document

=head2 published

    my $published = $norfcdoc->published;

Sets and returns the published date of the noRFC document

=head2 save

    $norfcdoc->save("myfile.txt") or die "Can't save file";

Save document to file

=head2 sha256

    my $sha256 = $norfcdoc->sha256;

Returns source file SHA-256 checksum of the noRFC document

=head2 size

    my $size = $norfcdoc->size;

Returns source file size of the noRFC document

=head2 title

    my $title = $norfcdoc->title;

Sets and returns the title of the noRFC document

=head2 updates

    my $updates = $norfcdoc->updates;

Sets and returns the updates of the noRFC document

=head2 wrap

    my $wrapped_text = $norfcdoc->wrap( $text );
    my $wrapped_body = $norfcdoc->wrap; # ->body

Returns wrapped document. See also L<Pod::Text>, L<Pod::Text::Termcap>, L<Text::Wrap>

=head1 SEE ALSO

L<NoRFC>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION /;
$VERSION = '1.00';

use Encode; # Encode::_utf8_on();

use Text::Tabs qw/expand/;
use Mojo::File qw/path/;

use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use CTK::Util qw/ bsave /;

use NoRFC::Util qw/fdate fbytes filesize sha256sum/;
use NoRFC::Const qw/PAGE_WIDTH/;

use constant {
    NORFC_SEPARATOR => "-" x 80,
    COLUMNS_MIN     => 54,
    COLUMNS_MAX     => 500,
    TABSTOP_DEFAULT => 8,
    TABSTOP_MAX     => 64,
};

sub new {
    my $class = shift;
    my %args = @_;

    my $self = bless {
            number      => $args{number} // '',
            category    => $args{category} // '',
            published   => $args{published} // '',
            author      => $args{author} // '',
            title       => $args{title} // '',
            updates     => '',
            obsoletes   => '',

            # File
            file        => undef,
            path        => undef,
            size        => 0,  # File size
            sha256      => '', # File sha256
            binary      => 0, # File is binary
            body        => '',

            # Wrap
            columns     => 0, # Wrap width
            tabstop     => 0, # Wrap tabstop
        }, $class;

    # Wrap options
    if (defined($args{columns}) && ($args{columns} >= COLUMNS_MIN and $args{columns} <= COLUMNS_MAX)) {
        $self->{columns} = $args{columns};
    } elsif (defined($args{columns})) {
        $self->{columns} = PAGE_WIDTH;
    }
    if (defined($args{tabstop}) && ($args{tabstop} >= 1 and $args{tabstop} <= TABSTOP_MAX)) {
        $self->{tabstop} = $args{tabstop};
    } elsif (defined($args{tabstop})) {
        $self->{tabstop} = TABSTOP_DEFAULT;
    }

    # Set updates and obsoletes
    $self->updates($args{updates}) if $args{updates};
    $self->obsoletes($args{obsoletes}) if $args{obsoletes};

    # Set body or file
    if (defined($args{body})) {
        # Set body first (text mode only)
        $self->body($args{body})
    } elsif (defined($args{file})) {
        $self->file($args{file})
    }

    return $self;
}
sub is_binary {
    my $self = shift;
    return $self->{binary};
}
sub number {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{number}) unless defined($val);
    $self->{number} = $val;
    return $self->{number};
}
sub category {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{category}) unless defined($val);
    $self->{category} = $val;
    return $self->{category};
}
sub published {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{published}) unless defined($val);
    $self->{published} = $val;
    return $self->{published};
}
sub author {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{author}) unless defined($val);
    $self->{author} = $val;
    return $self->{author};
}
sub title {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{title}) unless defined($val);
    $self->{title} = $val;
    return $self->{title};
}
sub body {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{body}) unless defined($val);
    Encode::_utf8_on($val);
    $self->{body} = $val;
    return $self->{body};
}
sub updates {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{updates}) unless defined($val);
    $self->{updates} = is_array($val) ? join(", ", sort {$a cmp $b} @$val) // '' : $val;
    return $self->{updates};
}
sub obsoletes {
    my $self = shift;
    my $val = shift;
    return fv2null($self->{obsoletes}) unless defined($val);
    $self->{obsoletes} = is_array($val) ? join(", ", sort {$a cmp $b} @$val) // '' : $val;
    return $self->{obsoletes};
}
sub file {
    my $self = shift;
    my $f = shift;
    return fv2null($self->{file}) unless defined($f);
    $self->{file} = $f;

    # Set file size and sha256 sum
    $self->{size} = filesize($f);
    $self->{sha256} = sha256sum($f);

    # Binary file?
    $self->{binary} = -B $f ? 1 : 0;

    # Set path
    $self->{path} = path($f);

    # Get body from file (for text file only)
    unless ($self->{binary}) {
        $self->body($self->{path}->slurp), # Get content
    }

    # Set data from file
    #Encode::_utf8_on($val);

    return $self->{file};
}
sub size {
    my $self = shift;
    return $self->{size}
}
sub sha256 {
    my $self = shift;
    return $self->{sha256}
}

sub as_string {
    my $self = shift;
    my $meta = shift;
       $meta = 1 if $self->is_binary;
    my $out = "";
    my $bodysize = length($self->body);

    # Header
    $out .= sprintf "Document    : %s\n", $self->number;
    $out .= sprintf "Category    : %s\n", $self->category;
    $out .= sprintf "Updates     : %s\n", $self->updates if $self->updates;
    $out .= sprintf "Obsoletes   : %s\n", $self->obsoletes if $self->obsoletes;
    $out .= sprintf "Published   : %s\n", is_int($self->published) ? fdate($self->published, 1) : $self->published;

    # Meta data
    if ($meta) {
        $out .= sprintf "File size   : %s\n", fbytes($self->size) if $self->size;
        $out .= sprintf "File sha256 : %s\n", $self->sha256 if $self->sha256;
        $out .= sprintf "Body size   : %s\n", fbytes($bodysize) if $bodysize;
    }
    $out .= sprintf "Author      : %s\n", $self->author;
    $out .= sprintf "Title       : %s\n", $self->title;

    # Content
    if ($bodysize) {
        #$out .= sprintf("%s\n", NORFC_SEPARATOR);
        $out .= "\n";
        $out .= $self->wrap;
    }

    Encode::_utf8_on($out);
    return $out;
}
sub save {
    my $self = shift;
    my $file = shift;
    return bsave($file, $self->as_string, 1);
}
sub wrap {
    my $self = shift;
    my $raw = shift // $self->body // '';
    return $raw unless length $raw;
    return $raw unless $self->{columns};
    $Text::Tabs::tabstop = $self->{tabstop} if $self->{tabstop};
    Encode::_utf8_on($raw);

    # Output string
    my @output = ();

    foreach my $l (_parse_document($raw)) {
        $l = expand($l) if $self->{tabstop};
        my $spaces = $l =~ /^([ \t]+)/ ? $1 : ''; # lead spaces
        $l =~ s/^([ \t]+)// if length($spaces);
        my $width = $self->{columns} - length($spaces);
        if (length($l) <= $width) {
            push @output, $spaces . $l;
            next;
        }
        while (length($l) > $width) {
            if ($l =~ s/^([^\n]{0,$width})[ \t\n]+// or $l =~ s/^([^\n]{$width})//) {
                push @output, $spaces . $1;
            } else {
                last;
            }
        }
        push @output, $spaces . $l if length $l;
    }

    return join "\n", @output;
}

sub _parse_document {
    my @lines;
    foreach my $lg (@_) {
        next unless defined $lg and length $lg;
        pos($lg) = 0;
        while($lg =~ /([^\n\r]*)(\r?\n?)/g) { # supports \r, \n ,\r\n
            #print(">$1<\n");
            push @lines, $1 if length($1) or length($2) or pos($lg) != length($lg);
            # I.e., unless it's a zero-length "empty line" at the very
            #  end of "foo\nbar\n" (i.e., between the \n and the EOS).
        }
    }
    return @lines;
}

1;

__END__
