package NoRFC::Progress;
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Progress - Show a progress bar on a standard terminal

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use NoRFC::Progress;

    use Time::HiRes qw(usleep);

    my $bar = NoRFC::Progress->new();
    $bar->start("Download... ", 1, 32);
    foreach my $i ($bar->min .. $bar->max) {

        usleep(500000); # ... too long operation ...

        $bar->update($i);
    }
    $bar->finish("Done");

=head1 DESCRIPTION

This module provides a progress bar on a standard terminal

=head2 new

    my $bar = NoRFC::Progress->new();
    my $bar = NoRFC::Progress->new(
            message     => 'Download... ',
            done        => 'Done',
            template    => "%.2f%%, %s left",
        );

Creates new object instance

=over 8

=item B<message>

Message of progress bar

=item B<done>

The message for finish

=item B<template>

Default of template

Examples:

    "%s%.2f%%, %s left"
    "%s%.2f%% completed, %s left"

Default: "%s%.2f%%, %s left"

=back

=head2 elapsed

    print $bar->elapsed;

Get elapsed time

=head2 finish

    $bar->finish("Done"); # Download... Done

Finishing the progress bar

=head2 max

    my $max = $bar->max; # 32

Returns the max value

=head2 min

    my $min = $bar->min; # 1

Returns the min value

=head2 start

    $bar->start($message, $min, $max);
    $bar->start("Download... ", 1, 32);

Start progress bar

=head2 restart

    $bar->restart;

Restart progress by with old data

=head2 update

    $bar->update(16); # Download... 50.00%, 11.5582 seconds left
    $bar->update("[file.dat] ", 16); # Download... [file.dat] 50.00%, 11.5582 seconds left

Update the progress bar

=head1 SEE ALSO

L<Progress::Any>, L<Term::ProgressBar>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use Time::HiRes ();
use NoRFC::Util qw/fduration/;

our $VERSION = "1.00";

use constant BAR_TEMPLATE => "%s%.2f%%, %s left";

sub new {
    my $class = shift;
    my %args = @_;

    my $self = bless {
            message     => $args{message} // '',
            done        => $args{done} // '',
            template    => $args{template} // BAR_TEMPLATE,
            started     => 0,
            started_at  => 0, # time
            finished_at => 0, # time
            elapsed     => 0, # time
            first_value => 0,
            last_value  => 0,
            oldlen      => 0,
            is_term     => (-t STDOUT) ? 1 : 0,
        }, $class;

    return $self;
}
sub start {
    my $self = shift;
    my $message = shift;
       $self->{message} = $message if defined $message;
       $message //= $self->{message} // '';
    my $first_value = shift;
       $self->{first_value} = $first_value if defined $first_value;
    my $last_value = shift;
       $self->{last_value} = $last_value if defined $last_value;

    $self->{started}   = 1;
    $self->{started_at} = Time::HiRes::time();
    $self->{finished_at} = 0;

    if (length $message) {
        print $message;
        print "\n" unless $self->{is_term};
    }

    return $self;
}
sub restart {
    my $self = shift;

    $self->{started}   = 1;
    $self->{finished_at} = 0;

    if (length $self->{message}) {
        print $self->{message};
        print "\n" unless $self->{is_term};
    }
}
sub update {
    my $self = shift;
    my $pos = pop || 0;
    my $pfx = shift // '';
    return $self unless $self->{is_term};

    local $| = 1;

    # PRC & ETA
    my $tm = Time::HiRes::time() - $self->{started_at};
    my $min = $self->min;
    my $max = $self->max;
    my $k = ($pos-1)/($max-$min);
    my $eta = $k ? fduration(($tm * (1-$k))/$k) : "unknown";

    # Result
    my $str = sprintf($self->{template}, $pfx, $k*100, $eta);
    my $str_len = length($str);
    my $old_len = $self->{oldlen} || 0;
    my $diff_len = ($old_len > $str_len) ? ($old_len - $str_len) : 0;
    $self->{oldlen} = $str_len;

    # Clear
    print "\b" x $old_len if $old_len;

    # Print
    print $str;
    print " " x $diff_len, "\b" x $diff_len if $diff_len;

    return $self;
}
sub finish {
    my $self = shift;
    my $done = shift // $self->{done} // '';

    local $| = 1;

    # Clear
    if ($self->{oldlen}) {
        print "\b" x $self->{oldlen};
        print " " x $self->{oldlen};
        print "\b" x $self->{oldlen};
    }

    # Print
    print $done;
    print "\n";

    $self->{oldlen} = 0;
    $self->{started} = 0;
    $self->{finished_at} = Time::HiRes::time();
    $self->{elapsed} = $self->{finished_at} - $self->{started_at};

    return $self;
}
sub min {
    my $self = shift;
    return $self->{first_value} || 0;
}
sub max {
    my $self = shift;
    return $self->{last_value} || 0;
}
sub elapsed {
    my $self = shift;
    my $el = $self->{started}
        ? Time::HiRes::time() - $self->{started_at}
        : $self->{finished_at} - $self->{started_at};
    return fduration($el);
}

1;

__END__

#print "\033[JStatus: 15% Completed."."\033[G";
#sleep 2;
#print "\033[JStatus: 20% Completed."."\033[G";
#print "\n";
