package NoRFC::Server;
use strict;
use warnings FATAL => 'all';
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Server - NoRFC web-server class

=head1 VERSION

Version 1.03

=head1 SYNOPSIS

    use NoRFC::Server;

=head1 DESCRIPTION

This module provides NoRFC web-server functionality

=head2 reload

The reload hook

=head2 startup

Mojolicious application startup method

=head1 SEE ALSO

L<Mojolicious>, L<NoRFC>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use Mojo::Base 'Mojolicious';

use Mojo::File qw/path/;
use Mojo::Util qw/sha1_sum secure_compare/;

use CTK::Util qw/preparedir sharedir sharedstatedir/;
use CTK::ConfGenUtil;
use CTK::TFVals qw/is_void/;

use NoRFC::Const qw/:GENERAL :SERVER :NUMBER/;
use NoRFC::Number;
use NoRFC::Util qw/explain human2bytes/;

#use NoRFC::Server::Auth;

use vars qw/ $VERSION /;
$VERSION = "1.03";

has 'documentroot'; # /usr/share/norfc
has 'dbdir'; # /var/lib/norfc/db
has 'maxstorefilesize'; # MaxStoreFileSize from config

# Mojo Routes (paths)
sub startup {
    my $self = shift; # app
    my $ctk = $self->can('ctk') ? $self->ctk() : $self->{ctk};

    # Define auth plugin
    $self->plugin('NoRFC::Server::Auth', {
            realm => "NoRFC Strict Zone",
            auth_render => {
                status => 401,
                json => {
                    status => 0,
                    message => "Authentication required!",
                },
            },
            fail_render => {
                status => 403,
                json => {
                    status => 0,
                    message => "Forbidden!",
                },
            },
            authen => sub {
                my ($c, $in_realm, $in_user, $in_pass, $params) = @_;
                my $credentials = $ctk->{credentials};
                return 1 if is_void($credentials);
                my $req_uri = $c->req->url->to_abs();
                my $in = sprintf("%s:%s",
                    $in_user // $req_uri->username() // "anonymous",
                    sha1_sum($in_pass // $req_uri->password() // time()),
                );
                #$c->log->warn(sprintf(">>> %s / %s", $in_realm, $in));
                foreach my $test (@$credentials) {
                    return 1 if secure_compare($test, $in);
                }
                return 0;
            },
        });

    # Logging
    $self->log->level("warn")->path($ctk->logfile()) unless $ctk->debugmode();

    # NoRFC Num object
    #$self->log->debug("Startup!! =$$"); # $self->ctk->logdir()
    my $norfcnum = NoRFC::Number->new();

    # Switch to installable home directory
    $self->home(Mojo::Home->new($ctk->datadir()));

    # Get DocumentRoot and replace as public-path
    my $documentroot = value($ctk->conf("documentroot")) || path(sharedir(), $ctk->prefix)->to_string();
    $self->documentroot($documentroot);
    $self->static->paths()->[0] = $documentroot; #unshift @{$static->paths}, '/home/sri/themes/blue/public';
    $self->static->paths()->[1] = $ctk->tempdir();

    # Database directory (for files)
    $self->dbdir(path($ctk->datadir, DBDIR)->to_string());

    # Get MaxStoreFileSize from config
    $self->maxstorefilesize(human2bytes(lvalue($ctk->config("maxstorefilesize")) // '64kB'));

    # Instance of database (reinit)
    my $store_args = $ctk->{store_args};
    $ctk->{store} = NoRFC::Store->new(%$store_args);

    # Hooks
    $self->hook(before_dispatch => sub {
        my $c = shift;

        # Set Server header
        $c->res->headers->server(sprintf("%s/%s", PROJECTNAME(), NoRFC->VERSION));
    });

    # Routes
    my $r = $self->routes;
    $r->get('/')->to('alpha#root');
    $r->get('/info')->to('alpha#info'); # Authentication inside (in test mode)
    $r->get('/info/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->to('alpha#info');
    $r->get('/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->to('alpha#norfc'); # Authentication inside
    $r->get('/search')->to('alpha#search');
    $r->post('/search')->to('alpha#search');
    $r->get('/rel/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->to('alpha#relget');

    # Routes with auth
    my %reqs = (authenticated => 1, authorized => 1);
    if (Mojolicious->VERSION < 9.00) {
        $r->get('/env')->over(%reqs)->to('alpha#env') if $ctk->debugmode();
        $r->get('/gendocnum')->over(%reqs)->to('alpha#gendocnum');
        $r->put('/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->over(%reqs)->to('alpha#publish');
        $r->delete('/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->over(%reqs)->to('alpha#remove');
        $r->post('/rel/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->over(%reqs)->to('alpha#reladd');
        $r->delete('/rel/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->over(%reqs)->to('alpha#reldel');
    } else {
        $r->get('/env')->requires(%reqs)->to('alpha#env') if $ctk->debugmode();
        $r->get('/gendocnum')->requires(%reqs)->to('alpha#gendocnum');
        $r->put('/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->requires(%reqs)->to('alpha#publish');
        $r->delete('/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->requires(%reqs)->to('alpha#remove');
        $r->post('/rel/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->requires(%reqs)->to('alpha#reladd');
        $r->delete('/rel/:rfcnum' => [rfcnum => NORFCNUM_REGEXP_V2S])->requires(%reqs)->to('alpha#reldel');
    }

    # Favicon
    delete $self->static->extra->{'favicon.ico'};

    return 1;
}

# Reload hook
sub reload {
    my $self = shift;
    $self->log->warn("Request for reload $$"); # $self->ctk->logdir()
    return 1; # 1 - ok; 0 - error :(
}

1;

__END__
