package NoRFC::Server::Auth;
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Server::Auth - The Mojolicious Plugin for authentication and authorization

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    # in your startup
    $self->plugin('NoRFC::Server::Auth', {
            realm => "NoRFC Strict Zone",
            auth_render => {
                status => 401,
                json => {
                    status => 0,
                    message => "Authentication required!",
                },
            },
            fail_render => {
                status => 403,
                json => {
                    status => 0,
                    message => "Forbidden!",
                },
            },
        });

    # in your routes
    sub index {
        my $self = shift;

        # Authentication
        return unless $self->is_authenticated({ render_by_fail => 1 });

        # Authorization
        return unless $self->is_authorized({ render_by_fail => 1 });

        $self->render(...);
    }

    # or as condition in your startup
    $self->routes->get('/info')->requires(authenticated => 1, authorized => 1)
        ->to('alpha#info');

    # or bridged in your startup
    my $auth = $self->routes->under(sub {
        my $self = shift;

        # Authentication
        return unless $self->is_authenticated({ render_by_fail => 1 });

        # Authorization
        return unless $self->is_authorized({ render_by_fail => 1 });

        return 1;
    });
    $auth->get('/info')->to('alpha#info');

=head1 DESCRIPTION

The Mojolicious Plugin for authentication and authorization

This plugin provides HTTP Basic Authentication and authorization

=head1 OPTIONS

This plugin supports the following options

=head2 authen

Authentication checker callback

    $self->plugin('NoRFC::Server::Auth', {authen => sub {
        my ($controller, $realm, $username, $password, $params) = @_;

        # ...

        return 1; # or 0 on fail
    }});

The B<$params> holds options from L</is_authenticated> call directly

=head2 authz

Authorization checker callback

    $self->plugin('NoRFC::Server::Auth', {authz => sub {
        my ($controller, $params) = @_;

        # ...

        return 1; # or 0 on fail
    }});

The B<$params> holds options from L</is_authorized> call directly

=head2 auth_render

Defines what is to be rendered when the authenticated condition is not met

Set to a coderef which will be called with the following signature:

    sub {
        my $controller = shift;
        my $realm = shift
        ...
        return $hashref;
    }

The return value of the subroutine will be ignored if it evaluates to false.
If it returns a hash reference, it will be dereferenced and passed as-is
to the controller's C<render> function

If set directly to a hash reference, that will be passed to C<render> instead

=head2 fail_render

Defines what is to be rendered when the authorized condition is not met

Set to a coderef which will be called with the following signature:

    sub {
        my $controller = shift;
        ...
        return $hashref;
    }

See also L</auth_render>

=head2 realm

    $self->plugin('NoRFC::Server::Auth', {realm => 'My Castle!'});

HTTP Realm, defaults to 'Strict Zone'

=head1 HELPERS

=head2 is_authenticated

This helper performs credential validation and checks the authentication status

    my $authenticated = $self->is_authenticated;
    my $authenticated = $self->is_authenticated({ render_by_fail => 1 });

The B<render_by_fail> enables rendering the fail response. See L</auth_render>

=head2 is_authorized

This helper checks the authorization status

    my $authorized = $self->is_authorized;
    my $authorized = $self->is_authorized({ render_by_fail => 1 });

The B<render_by_fail> enables rendering the fail response. See L</fail_render>

=head1 METHODS

Internal methods

=head2 register

This method register the plugin and helpers in L<Mojolicious> application.

=head1 EXAMPLES

Examples of using

=head2 ROUTING VIA CONDITION

This plugin exports a routing condition you can use in order to limit
access to certain documents to only authenticated users.

    $self->routes->get('/info')->requires(authenticated => 1, authorized => 1)
        ->to('alpha#info');

Prior to Mojolicious 9, use "over" instead of "requires."

=head2 ROUTING VIA CALLBACK

If you want to be able to send people to a login page, you will have to use
the following:

    sub index {
        my $self = shift;

        $self->redirect_to('/login') and return 0
            unless($self->is_authenticated && $self->is_authorized);

        $self->render(...);
    }

=head2 ROUTING VIA BRIDGE

    my $auth = $self->routes->under(sub {
        my $self = shift;

        # Authentication
        return unless $self->is_authenticated({ render_by_fail => 1 });

        # Authorization
        return unless $self->is_authorized({ render_by_fail => 1 });

        return 1;
    });
    $auth->get('/info')->to('alpha#info');

=head1 SEE ALSO

L<Mojolicious>, L<Mojolicious::Plugin::Authentication>, L<Mojolicious::Plugin::Authorization>,
L<Mojolicious::Plugin::BasicAuth>, L<Mojolicious::Plugin::HttpBasicAuth>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use Mojo::Base 'Mojolicious::Plugin';
use Mojo::Util qw/b64_decode/;

our $VERSION = '1.00';

sub register {
    my ($plugin, $app, $defaults) = @_; # $self = $plugin
    $defaults //= {};
    $defaults->{realm} //= 'Strict Zone';
    $defaults->{fail_render_cb} = ref($defaults->{fail_render}) eq 'CODE'
        ? $defaults->{fail_render}
        : sub { return $defaults->{fail_render} };
    $defaults->{auth_render_cb} = ref($defaults->{auth_render}) eq 'CODE'
        ? $defaults->{auth_render}
        : sub {
            my $c = shift; # $controller
            my $realm = shift || $defaults->{realm};
            $c->res->headers->www_authenticate(sprintf('Basic realm="%s"', $realm));
            return $defaults->{auth_render}
        };

    # Authentication condition + fail render
    $app->routes->add_condition(authenticated => sub {
        my ($r, $c, $captures, $required) = @_;

        #$c->log->warn("!!! condition=authenticated !!!");

        my $res = (!$required or $c->is_authenticated);
        unless ($res) {
            my $render = $defaults->{auth_render_cb}; # Code
            my $fail = $render->($c); # Call render, returns {}
            $c->render(%$fail) if $fail;
        }
        return $res;
    });

    # Authorization condition + fail render
    $app->routes->add_condition(authorized => sub {
        my ($r, $c, $captures, $required) = @_;

        #$c->log->warn("!!! condition=authorized !!!");

        my $res = (!$required or $c->is_authorized);
        unless ($res) {
            my $render = $defaults->{fail_render_cb}; # Code
            my $fail = $render->($c); # Call render, returns {}
            $c->render(%$fail) if $fail;
        }
        return $res;
    });


    # Authentication checker (authen)
    $app->helper(is_authenticated => sub {
        my $c = shift;
        my $params = shift // {};
        my %opt = (%$defaults, %$params); # Hashes merging

        # Define the authen callback
        my $authen = ref($opt{authen}) eq 'CODE' ? $opt{authen} : sub { 1 };

        # Get authorization string from request headers
        my $auth_string = $c->req->headers->authorization
            || $c->req->env->{'X_HTTP_AUTHORIZATION'} || $c->req->env->{'HTTP_AUTHORIZATION'} || '';
        if ($auth_string =~ /Basic\s+(.*)/) {
            $auth_string = $1;
        }
        my $auth_pair = b64_decode($auth_string);

        #$c->log->warn("!!! call is_authenticated");
        #$c->log->warn("    auth_pair=" . $auth_pair);

        # Verification
        return 1 if $auth_pair && $authen->($c, $opt{realm}, split(/:/, $auth_pair, 2), $params);
           # $controller, $realm, $username, $password, $params

        # Render by fail
        if ($opt{render_by_fail}) {
            #$c->log->warn("!!! render_by_fail");
            my $render = $opt{auth_render_cb}; # Code
            my $fail = $render->($c, $opt{realm}); # Call render, returns {}
            $c->render(%$fail) if $fail;
        }

        # Not authenticated
        return 0;
    });

    # Authorization checker (authz)
    $app->helper(is_authorized => sub {
        my $c = shift;
        my $params = shift // {};
        my %opt = (%$defaults, %$params); # Hashes merging

        # Define the authz callback
        my $authz = ref($opt{authz}) eq 'CODE' ? $opt{authz} : sub { 1 };

        #$c->log->warn("!!! is_authorized");

        # Verification
        return 1 if $authz->($c, $params); # $controller, $params

        # Render by fail
        if ($opt{render_by_fail}) {
            #$c->log->warn("!!! render_by_fail");
            my $render = $opt{fail_render_cb}; # Code
            my $fail = $render->($c); # Call render, returns {}
            $c->render(%$fail) if $fail;
        }

        # Not authorized
        return 0;
    });
}

1;

__END__
