package NoRFC::Store;
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Store - NoRFC store class

=head1 VERSION

Version 1.05

=head1 SYNOPSIS

    use NoRFC::Store;

    my $store = NoRFC::Store->new(
        file => "/tmp/test.db",
        attributes => "RaiseError=0; PrintError=0; sqlite_unicode=1",
    );

    my $store = NoRFC::Store->new(
        dsn => "DBI:mysql:database=norfc;host=mysql.example.com",
        user => "username",
        password => "password",
        set => [
            "RaiseError        0",
            "PrintError        0",
            "mysql_enable_utf8 1",
        ],
    );

    die($store->error) unless $store->status;

=head1 DESCRIPTION

This module provides store methods.

=head2 SQLITE DDL

    CREATE TABLE IF NOT EXISTS "norfc" (
      "id"          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
      "num"         CHAR(32) NOT NULL UNIQUE, -- Document number
      "pub"         INTEGER NOT NULL, -- Public date (unixtime)
      "category"    CHAR(64) DEFAULT NULL, -- Cetegory of document
      "title"       TEXT DEFAULT NULL, -- Title of document
      "author"      CHAR(255) DEFAULT NULL, -- Author of document
      "attributes"  CHAR(255) DEFAULT NULL, -- Attributes
      "content"     TEXT DEFAULT NULL, -- Source content
      "signature"   TEXT DEFAULT NULL -- Signature
    );
    CREATE UNIQUE INDEX "num" ON "norfc" (
      "num"
    );
    CREATE TABLE IF NOT EXISTS "rels" (
      "id"          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
      "num"         CHAR(32) NOT NULL, -- Document number
      "ext"         HAR(32) NOT NULL, -- External document number, link
      "name"        CHAR(64) DEFAULT NULL
    );

=head2 MYSQL DDL

    CREATE DATABASE `norfc` /*!40100 DEFAULT CHARACTER SET utf8 COLLATE utf8_bin */;
    CREATE TABLE IF NOT EXISTS `norfc` (
      `id`          INT(11) NOT NULL AUTO_INCREMENT,
      `num`         VARCHAR(32) NOT NULL, -- Document number
      `pub`         INT(11) NOT NULL, -- Unixtime public date
      `category`    varchar(64) COLLATE utf8_bin NOT NULL, -- Category name
      `title`       TEXT COLLATE utf8_bin NOT NULL, -- Title
      `author`      VARCHAR(255) DEFAULT NULL,
      `attributes`  VARCHAR(255) DEFAULT NULL,
      `content`     LONGTEXT COLLATE utf8_bin DEFAULT NULL,
      `signature`   TEXT COLLATE utf8_bin DEFAULT NULL,
      PRIMARY KEY (`id`),
      UNIQUE KEY `num` (`num`)
    ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin;
    CREATE TABLE IF NOT EXISTS `rels` (
      `id`          INT(11) NOT NULL AUTO_INCREMENT,
      `num`         VARCHAR(32) NOT NULL, -- Document number
      `ext`         VARCHAR(32) NOT NULL, -- External document number, link
      `name`        VARCHAR(64) DEFAULT NULL,
      PRIMARY KEY (`id`)
    ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin;

=head2 POSTGRESQL DDL

    CREATE TABLE IF NOT EXISTS `norfc` (
      `id`          INT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
      `num`         CHAR(32) NOT NULL UNIQUE, -- Document number
      `pub`         INTEGER NOT NULL, -- Unixtime public date
      `category`    CHAR(64) DEFAULT NULL, -- Category name
      `title`       TEXT DEFAULT NULL, -- Title
      `author`      CHAR(255) DEFAULT NULL,
      `attributes`  CHAR(255) DEFAULT NULL,
      `content`     TEXT COLLATE utf8_bin DEFAULT NULL,
      `signature`   TEXT COLLATE utf8_bin DEFAULT NULL
    );
    CREATE TABLE IF NOT EXISTS `rels` (
      `id`          INT GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY,
      `num`         CHAR(32) NOT NULL, -- Document number
      `ext`         CHAR(32) NOT NULL, -- External document number, link
      `name`        CHAR(64) DEFAULT NULL
    );

=head1 METHODS

=head2 new

    my $store = NoRFC::Store->new(
        file => "/tmp/test.db",
        attributes => "RaiseError=0; PrintError=0; sqlite_unicode=1",
    );

    # ... or MySQL:

    my $store = NoRFC::Store->new(
        dsn => "DBI:mysql:database=norfc;host=mysql.example.com",
        user => "username",
        password => "password",
        compress => 1,
        set => [
            "RaiseError        0",
            "PrintError        0",
            "mysql_enable_utf8 1",
        ],
    );

Creates DBI object

=head2 add

    $store->add(
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document",
        author      => "Serz Minus",
        content     => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Add new recored

=head2 count

    print $store->count();

Returns count of records

=head2 del

    $store->del("RFC-R0000000") or die($store->error);

Delete record by document number

    $store->del(1) or die($store->error);

Delete record by record id

=head2 dsn

    my $dsn = $store->dsn;

Returns DSN string of current database connection

=head2 error

    my $error = $store->error;

Returns error message

    my $status = $store->error( "Error message" );

Sets error message if argument is provided.
This method in "set" context returns status of the operation as status() method.

=head2 get

    my %data = $store->get("RFC-R0000000");

Returns data from database by document number

    my %data = $store->get(1);

Returns data from database by record id

=head2 getinfo

    my %data = $store->get("RFC-R0000000");

Returns only info data from database by document number

=head2 getall

    my @table = $store->getall();
    my @table_100 = $store->getall(100); # row_count
    my @table_100 = $store->getall(100, 100); # offset, row_count

Returns pure data from database with limit supporting

=head2 getid

    my $id = $store->getid("RFC-R0000000");

Returns id of record from database by document number

=head2 is_sqlite

    print $store->is_sqlite ? "Is SQLite" : "Is NOT SQLite"

Returns true if type of current database is SQLite

=head2 ping

    $store->ping ? 'OK' : 'Database session is expired';

Checks the connection to database

=head2 reladd

    my $id = $store->reladd(
        number      => "NoRFC-R202300913",
        external    => "NoRFC-R202300952",
        name        => "Unknown",
    ) or die($store->error);

Add new relation and returns ID

=head2 reldel

    my $status = $store->reldel("NoRFC-R202300913");
    my $status = $store->reldel("NoRFC-R202300913", "NoRFC-R202300952");

Removes relations of the noRFC document number and returns status

=head2 relget

    my @table = $store->relget("NoRFC-R202300913");
    my @table = $store->relget("NoRFC-R202300913", "NoRFC-R202300952");

Returns relations of the noRFC document number

=head2 replace

    $store->replace(
        id          => 123,
        num         => "NoRFC-R202300963",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document (edited)",
        author      => "Serz Minus",
        content     => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Replace pure recored by id

=head2 search

    my @table = $store->search(
        offset      => 0,
        row_count   => 10,
        category    => "Etcetera",
        title       => "title",
    );

Search documents by category and title fragment

    my ($count) = $store->search(
        get_count_only => 1,
        category    => "Etcetera",
        title       => "title",
    );

Search documents by category and title fragment and returns number of found documents

=head2 set

    $store->set(
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document (edited)",
        author      => "Serz Minus",
        content     => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Update recored by document number

    $store->set(
        id          => 1,
        num         => "RFC-R0000000",
        pub         => 1590000000,
        category    => "Unknown",
        title       => "Test document (edited)",
        author      => "Serz Minus",
        content     => "...Content of the document...",
        signature   => "***",
    ) or die($store->error);

Update recored by record id

=head2 status

    my $status = $store->status;
    my $status = $store->status( 1 ); # Sets the status value and returns it

Gets or sets the BOOL status of the operation

=head2 truncate

    $store->truncate or die($store->error);

Delete all records from tables

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 BUGS

* none noted

=head1 SEE ALSO

L<CTK::DBI>, L<NoRFC>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/$VERSION/;
$VERSION = '1.05';

use Carp;
use CTK::DBI;
use CTK::Util qw/ touch dformat /;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use File::Spec;
use Encode qw/encode decode/;
use MIME::Base64 qw/encode_base64 decode_base64/;
use Compress::Zlib qw/compress uncompress/;
use NoRFC::Util qw/attr2hash hash2attr/;

use constant {
    DEFAULT_DSN_MASK    => 'DBI:SQLite:dbname=%s',
    DEFAULT_DBI_ATTR    => {
            dsn         => '', # See DEFAULT_DSN_MASK
            user        => '',
            password    => '',
            set         => [
                    'RaiseError 0',
                    'PrintError 0',
                    'sqlite_unicode 1',
                ],
        },
};

use constant DDL_CREATE_NORFC => <<'DDL';
CREATE TABLE IF NOT EXISTS "norfc" (
    "id"        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
    "num"       CHAR(32) NOT NULL UNIQUE, -- Document number
    "pub"       INTEGER NOT NULL, -- Public date (unixtime)
    "category"  CHAR(64) DEFAULT NULL, -- Cetegory of document
    "title"     TEXT DEFAULT NULL, -- Title of document
    "author"    CHAR(255) DEFAULT NULL, -- Author of document
    "attributes" CHAR(255) DEFAULT NULL, -- Attributes
    "content"   TEXT DEFAULT NULL, -- Source content
    "signature" TEXT DEFAULT NULL -- Signature
)
DDL
use constant DDL_CREATE_INDEX => <<'DDL';
CREATE UNIQUE INDEX "num" ON "norfc" (
    "num"
)
DDL
use constant DDL_CREATE_RELS => <<'DDL';
CREATE TABLE IF NOT EXISTS "rels" (
    "id"        INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
    "num"       CHAR(32) NOT NULL, -- Document number
    "ext"       CHAR(32) NOT NULL, -- External document number, link
    "name"      CHAR(64) DEFAULT NULL
)
DDL

# INSERT INTO norfc
#    (`num`,`pub`,`category`,`title`,`author`,`attributes`, `content`,`signature`)
# VALUES
#    ("RFC-R0000000", 1590000000, "Unknown", "Test document", "Serz Minus", "compress=0", "...Content of the document...", "***")
use constant DML_ADD => <<'DML';
INSERT INTO `norfc`
    (`num`,`pub`,`category`,`title`,`author`,`attributes`, `content`,`signature`)
VALUES
    (?,?,?,?,?,?,?,?)
DML

use constant DML_PURE_ADD => <<'DML';
INSERT INTO `norfc`
    (`id`,`num`,`pub`,`category`,`title`,`author`,`attributes`,`content`,`signature`)
VALUES
    (?,?,?,?,?,?,?,?,?)
DML

use constant DML_GET_COUNT => <<'DML';
SELECT COUNT(`id`) AS `cnt`
FROM `norfc`
DML

use constant DML_GET_BY_ID => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`attributes`,`content`,`signature`
FROM `norfc`
WHERE `id` = ?
DML

use constant DML_GET_INFO_BY_NUM => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`attributes`,`signature`
FROM `norfc`
WHERE `num` = ?
DML

use constant DML_GET_ID_BY_NUM => <<'DML';
SELECT `id`
FROM `norfc`
WHERE `num` = ?
DML

use constant DML_GET_BY_NUM => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`attributes`,`content`,`signature`
FROM `norfc`
WHERE `num` = ?
DML

use constant DML_GETALL => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`,`author`,`attributes`,`content`,`signature`
FROM `norfc`
ORDER BY `pub` ASC
DML

use constant DML_SEARCH => <<'DML';
SELECT `id`,`num`,`pub`,`category`,`title`
FROM `norfc` [CONDITION]
ORDER BY `pub` ASC
[LIMIT]
DML

use constant DML_SEARCH_COUNT => <<'DML';
SELECT COUNT(`id`) AS `cnt`
FROM `norfc` [CONDITION]
DML

use constant DML_DEL_BY_ID => <<'DML';
DELETE FROM `norfc` WHERE `id` = ?
DML

use constant DML_DEL_BY_NUM => <<'DML';
DELETE FROM `norfc` WHERE `num` = ?
DML

use constant DML_TRUNCATE_NORFC => <<'DML';
DELETE FROM `norfc`
DML
use constant DML_TRUNCATE_RELS => <<'DML';
DELETE FROM `rels`
DML

use constant DML_SET_BY_ID => <<'DML';
UPDATE `norfc`
SET `num` = ?, `pub` = ?, `category` = ?, `title` = ?, `author` = ?, `attributes` = ?, `content` = ?, `signature` = ?
WHERE `id` = ?
DML

use constant DML_SET_BY_NUM => <<'DML';
UPDATE `norfc`
SET `num` = ?, `pub` = ?, `category` = ?, `title` = ?, `author` = ?, `attributes` = ?, `content` = ?, `signature` = ?
WHERE `num` = ?
DML

#
# Relations
#

use constant DML_REL_ADD => <<DML;
INSERT INTO
    `rels` (`num`,`ext`,`name`)
VALUES
    (?, ?, ?)
DML

use constant DML_REL_GET_INFO => <<DML;
SELECT `id`,`num`,`ext`,`name`
FROM `rels`
WHERE `num` = ? AND `ext` = ?
DML

use constant DML_REL_GET_BY_NUMBER => <<DML;
SELECT `id`,`num`,`ext`,`name`
FROM `rels`
WHERE `num` = ?
DML

use constant DML_REL_DEL_BY_NUMBERS => <<DML;
DELETE FROM `rels`
WHERE `num` = ? AND `ext` = ?
DML

use constant DML_REL_DEL_BY_NUMBER => <<DML;
DELETE FROM `rels`
WHERE `num` = ?
DML

sub new {
    my $class = shift;
    my %args = @_;
    my $args_set = $args{'set'};
    unless ($args{dsn}) {
        my $dda = DEFAULT_DBI_ATTR;
        foreach (keys %$dda) {
            next if $_ eq 'set';
            $args{$_} //= $dda->{$_}
        }
    }
    my $file = $args{file} // "";
    my $dsn = $args{dsn} || sprintf(DEFAULT_DSN_MASK, $file);
    my %attrs_from_set = _set2attr($args_set);
    my %attrs_from_str = $args{attributes} ? attr2hash($args{attributes}) : ();
    my %attrs = (%attrs_from_set, %attrs_from_str);
       %attrs = _set2attr(DEFAULT_DBI_ATTR()->{set}) unless %attrs;

    # DB
    my $db = CTK::DBI->new(
        -dsn        => $dsn,
        -debug      => 0,
        -username   => $args{'user'},
        -password   => $args{'password'},
        -attr       => {%attrs},
        $args{timeout} ? (
            -timeout_connect => $args{timeout},
            -timeout_request => $args{timeout},
        ) : (),
    );
    my $dbh = $db->connect if $db;

    # SQLite
    my $fnew = 0;
    my $issqlite = 0;
    if ($dbh && $dsn =~ /SQLite/i) { # If SQLite (default)
        $file = $dbh->sqlite_db_filename();
        unless ($file && (-e $file) && !(-z $file)) {
            touch($file);
            chmod(0666, $file);
            $fnew = 1;
        }
        $issqlite = 1;
    }

    # Defaults
    my $status = 1; # Ok
    my $error = "";
    if (!$db) {
        $error = sprintf("Can't init database \"%s\"", $dsn);
        $status = 0;
    } elsif (!$dbh) {
        $error = sprintf("Can't connect to database \"%s\": %s", $dsn, $DBI::errstr || "unknown error");
        $status = 0;
    } elsif ($fnew) {
        $db->execute(DDL_CREATE_NORFC);
        $db->execute(DDL_CREATE_INDEX) unless $dbh->err;
        $db->execute(DDL_CREATE_RELS) unless $dbh->err;
        $error = $dbh->errstr();
        $status = 0 if $dbh->err;
    }
    unless ($error) { # No errors
        unless ($dbh->ping) {
            $error = sprintf("Can't init database \"%s\". Ping failed: %s",
                $dsn, $dbh->errstr() || "unknown error");
            $status = 0;
        }
    }

    my $self = bless {
            file    => $file,
            issqlite=> $issqlite,
            compress=> $args{compress},
            dsn     => $dsn,
            error   => $error // "",
            dbi     => $db,
            status  => $status,
        }, $class;

    return $self;
}

sub status {
    my $self = shift;
    my $value = shift;
    return fv2zero($self->{status}) unless defined($value);
    $self->{status} = $value ? 1 : 0;
    return $self->{status};
}
sub error {
    my $self = shift;
    my $value = shift;
    return uv2null($self->{error}) unless defined($value);
    $self->{error} = $value;
    $self->status($value ne "" ? 0 : 1);
    return $value;
}
sub ping {
    my $self = shift;
    return 0 unless $self->{dsn};
    my $dbi = $self->{dbi};
    return 0 unless $dbi;
    my $dbh = $dbi->{dbh};
    return 0 unless $dbh;
    return 0 unless $dbh->can('ping');
    return $dbh->ping();
}
sub dsn {
    my $self = shift;
    return $self->{dsn};
}
sub is_sqlite {
    my $self = shift;
    return $self->{issqlite} ? 1 : 0;
}

# CRUD Methods

sub add {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Get attributes
    my %attr = attr2hash($data{attributes});
       $attr{compress} = $self->{compress} ? 1 : 0;
       delete $attr{compress} unless $attr{compress};

    # Encoding
    $data{content} //= '';
    Encode::_utf8_on($data{content}) unless $attr{binary} || $attr{encrypted};

    # Add
    $dbi->execute(DML_ADD,
        $data{num},
        $data{pub} || time(),
        $data{category},
        $data{title},
        $data{author},
        hash2attr(%attr),
        length($data{content}) ? encode_base64(
            $attr{compress}
                ? ($attr{binary} || $attr{encrypted})
                    ? compress($data{content}) # Binary/Encrypted
                    : compress(encode("UTF-8", $data{content}))
                : ($attr{binary} || $attr{encrypted})
                    ? $data{content} # Binary/Encrypted
                    : encode("UTF-8", $data{content})
        ) : '',
        $data{signature},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't insert new record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return 1;
}
sub count {
    my $self = shift;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my $cnt = $dbi->field(DML_GET_COUNT) || 0;
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get count: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }

    return $cnt;
}
sub get {
    my $self = shift;
    my $id = shift || 0;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my %rec = $dbi->recordh(is_int($id) ? DML_GET_BY_ID : DML_GET_BY_NUM, $id);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
        return ();
    }

    # Get attributes
    my %attr = attr2hash($rec{attributes});
    if ($attr{compress}) {
        $rec{content} = ($attr{binary} || $attr{encrypted})
            ? uncompress(decode_base64($rec{content})) # Binary/Encrypted
            : decode("UTF-8", uncompress(decode_base64($rec{content})));
    } else {
        $rec{content} = ($attr{binary} || $attr{encrypted})
            ? decode_base64($rec{content}) # Binary/Encrypted
            : decode("UTF-8", decode_base64($rec{content}));
    }

    return %rec;
}
sub getinfo {
    my $self = shift;
    my $num = shift || 0;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my %rec = $dbi->recordh(DML_GET_INFO_BY_NUM, $num);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
        return ();
    }

    return %rec;
}
sub getid {
    my $self = shift;
    my $num = shift;
    return 0 unless $num;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    my $id = $dbi->field(DML_GET_ID_BY_NUM, $num);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get id field: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }

    return $id;
}
sub getall {
    my $self = shift;
    my $row_count = pop;
    my $offset = pop;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};
    my $limit = $row_count
        ? $offset
            ? sprintf(" LIMIT %d, %d", $offset, $row_count)
            : sprintf(" LIMIT %d", $row_count)
        : "";

    my @tbl = $dbi->table(sprintf("%s%s", DML_GETALL,  $limit ));
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get records: %s", uv2null($dbi->connect->errstr)));
        return ();
    }
    return @tbl;
}
sub search {
    my $self = shift;
    my %in = @_;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Limit
    my $row_count = $in{row_count} || 0;
    my $offset = $in{offset} || 0;;
    my $limit = $row_count
        ? $offset
            ? sprintf("LIMIT %d, %d", $offset, $row_count)
            : sprintf("LIMIT %d", $row_count)
        : "";

    # Condition
    my @cond = ();
    if (defined $in{category} and length $in{category}) {
        push @cond, sprintf("`category` = %s", $dbi->{dbh}->quote($in{category}));
    }
    if (defined $in{title} and length $in{title}) {
        push @cond, sprintf("`title` LIKE %s", $dbi->{dbh}->quote(sprintf("%%%s%%", $in{title})));

    }
    my $condition = join(" AND ", @cond);

    # Count
    if ($in{get_count_only}) {
        my $cnt = $dbi->field(dformat(DML_SEARCH_COUNT, {
            CONDITION => $condition ? sprintf("WHERE %s", $condition) : "",
        }));
        if ($dbi->connect->err) {
            $self->error(sprintf("Can't get count: %s", uv2null($dbi->connect->errstr)));
            return ();
        }
        return ($cnt);
    }

    # Search
    my @tbl = $dbi->table(dformat(DML_SEARCH, {
        LIMIT => $limit,
        CONDITION => $condition ? sprintf("WHERE %s", $condition) : "",
    }));
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get records: %s", uv2null($dbi->connect->errstr)));
        return ();
    }
    return @tbl;
}
sub del {
    my $self = shift;
    my $id = shift || 0;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    $dbi->execute(is_int($id) ? DML_DEL_BY_ID : DML_DEL_BY_NUM, $id);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't delete record: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }
    return 1;
}
sub truncate {
    my $self = shift;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    $dbi->execute(DML_TRUNCATE_NORFC);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't truncate table: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }

    $dbi->execute(DML_TRUNCATE_RELS);
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't truncate table: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }

    return 1;
}
sub set {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Get attributes
    my %attr = attr2hash($data{attributes});
       $attr{compress} = $self->{compress} ? 1 : 0;
       delete $attr{compress} unless $attr{compress};

    # Encoding
    $data{content} //= '';
    Encode::_utf8_on($data{content}) unless $attr{binary} || $attr{encrypted};

    # Set by id or num
    $dbi->execute($data{id} ? DML_SET_BY_ID : DML_SET_BY_NUM,
        $data{num},
        $data{pub} || time(),
        $data{category},
        $data{title},
        $data{author},
        hash2attr(%attr),
        length($data{content}) ? encode_base64(
            $attr{compress}
                ? ($attr{binary} || $attr{encrypted})
                    ? compress($data{content}) # Binary/Encrypted
                    : compress(encode("UTF-8", $data{content}))
                : ($attr{binary} || $attr{encrypted})
                    ? $data{content} # Binary/Encrypted
                    : encode("UTF-8", $data{content})
        ) : '',
        $data{signature},
        $data{id} || $data{num},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't update record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return 1;
}
sub replace {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};
    my $id = $data{id};
    unless ($id) {
        $self->error("No id specified");
        return;
    }

    # Del
    $self->del($id) or return;

    # Add
    $dbi->execute(DML_PURE_ADD,
        $id, $data{num}, $data{pub}, $data{category}, $data{title},
        $data{author}, $data{attributes}, $data{content}, $data{signature},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't insert new record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return 1;
}
sub reladd {
    my $self = shift;
    my %data = @_;
    return unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Del
    $dbi->execute(DML_REL_DEL_BY_NUMBERS, $data{number}, $data{external});
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't delete record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    # Add
    $dbi->execute(DML_REL_ADD,
        $data{number},
        $data{external},
        $data{name},
    );
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't insert new record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    # Get id
    my %rec = $dbi->recordh(DML_REL_GET_INFO, $data{number}, $data{external});
    if ($dbi->connect->err) {
        $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
        return;
    }

    return $rec{id} || 0;
}
sub relget {
    my $self = shift;
    my $num = shift;
    my $ext = shift;
    return () unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};
    my @found = ();

    # Search by ext
    if ($num && $ext) {
        my %rec = $dbi->recordh(DML_REL_GET_INFO, $num, $ext);
        if ($dbi->connect->err) {
            $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
            return ();
        }
        push @found, {
            id      => $rec{id},
            number  => $rec{num},
            external=> $rec{ext},
            name    => $rec{name},
        };
    } elsif ($num) {
        my @tbl = $dbi->table(DML_REL_GET_BY_NUMBER, $num);
        if ($dbi->connect->err) {
            $self->error(sprintf("Can't get record: %s", uv2null($dbi->connect->errstr)));
            return ();
        }
        foreach my $rec (@tbl) {
            push @found, {
                id      => $rec->[0],
                number  => $rec->[1],
                external=> $rec->[2],
                name    => $rec->[3],
            };
        }
    }

    return @found;
}
sub reldel {
    my $self = shift;
    my $num = shift;
    my $ext = shift;
    return 0 unless $self->ping;
    $self->error("");
    my $dbi = $self->{dbi};

    # Delete by ext
    if ($num && $ext) {
        $dbi->execute(DML_REL_DEL_BY_NUMBERS, $num, $ext);
    } elsif ($num) {
        $dbi->execute(DML_REL_DEL_BY_NUMBER, $num);
    }

    if ($dbi->connect->err) {
        $self->error(sprintf("Can't delete record: %s", uv2null($dbi->connect->errstr)));
        return 0;
    }
    return 1;
}

sub _set2attr {
    my $in = shift;
    my $attr = is_array($in) ? $in : array($in => "set");
    my %attrs;
    foreach (@$attr) {
        $attrs{$1} = $2 if $_ =~ /^\s*(\S+)\s+(.+)$/;
    }
    return %attrs;
}

1;

__END__
