package NoRFC::Util;
use strict;
use utf8;

=encoding utf8

=head1 NAME

NoRFC::Util - NoRFC utilities

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    use NoRFC::Util;

=head1 DESCRIPTION

Exported utility functions

=head2 attr2hash

    my %hash = attr2hash("baz=; bar=123; foo=test");

Convert a string of attributes to hash

Example: foo=123, bar=gii,baz;qux=werwe

=head2 blue, cyan, green, red, yellow, magenta

    print cyan("Format %s", "text");

Returns colored string

=head2 explain

    print explain( $object );

Returns Data::Dumper dump

=head2 fbytes

    print fbytes( 123456 );

Returns formatted size value

=head2 fdate

    print fdate( time );
    print fdate( time, 1 ); # GMT

Returns formatted date value

=head2 fduration

    print fduration( 123 );

Returns formatted duration value

=head2 filesize

    my $fsize = filesize( $file );

Returns the file size

=head2 hash2attr

    my $attr = hash2attr(foo => 'test', bar => 123, baz => undef);
        # baz=; bar=123; foo=test

Convert a simple hash to string of attributes

=head2 human2bytes

    my $bytes = human2bytes("100 kB");

Converts a human readable byte count into the pure  number of bytes without any suffix

=head2 m11r_checkdigit

    my $m11r = m11r_checkdigit("123456789"); # 5

Returns M11R checkdigit by specified digits-string

See also L<Algorithm::CheckDigits::M11_015> and check_okpo()

=head2 md5sum

    my $md5 = md5sum( $file );

See L<Digest::MD5>

=head2 shasum

    my $sha1 = shasum( $file );
    my $sha256 = shasum( $file, 256 );

This function provides calculate SHA-1, SHA-224, SHA-256, SHA-384, SHA-512,
SHA-512/224, and SHA-512/256 digests. The function returns the digest encoded
as a hexadecimal string

Allowed values for algorithm (second optional param) are 1, 224, 256, 384, 512, 512224, or 512256.
It's also possible to use common string representations of the algorithm
(e.g. "sha256", "SHA-384"). If the argument is missing, SHA-1 will be used by default.

See L<Digest::SHA>

=head2 sha1sum

    my $sha1 = sha1sum( $file );

Calculates a SHA-1 sum of file

=head2 sha256sum

    my $sha256 = sha256sum( $file );

Calculates a SHA-256 sum of file

=head2 parse_credentials

    my ($user, $password) = parse_credentials( 'http://user:password@example.com' );
    my ($user, $password) = parse_credentials( new URI('http://user:password@example.com') );

Returns credentials pair by URL or URI object

=head2 parsewords

    my @b = parsewords("foo,bar baz"); # qw/foo bar baz/

Parses string and split it by words. See L<Text::ParseWords/quotewords>

=head1 HISTORY

See C<Changes> file

=head1 TO DO

See C<TODO> file

=head1 SEE ALSO

L<Algorithm::CheckDigits>, L<Digest::MD5>, L<Digest::SHA>, L<URI>, L<POSIX>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<http://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION @EXPORT_OK @EXPORT /;
$VERSION = '1.00';

use Carp;
use Term::ANSIColor qw/colored/;
use Text::ParseWords qw/quotewords shellwords/;
use Digest::MD5;
use Digest::SHA;
use URI;
use URI::Escape qw/uri_unescape/;
use POSIX qw/ceil/;

use NoRFC::Const qw/
        IS_TTY NORFCNUM_DIGITS_V1 NORFCNUM_REGEXP_V2 NORFCNUM_DIGITS_V2 NORFCNUM_REGEXP_V1
        DATETIME_FORMAT DATETIME_GMT_FORMAT
    /;

use CTK::Util qw/dtf tz_diff/;

use base qw/Exporter/;
@EXPORT = (qw/
        blue green red yellow cyan magenta
    /);
@EXPORT_OK = (qw/
        explain
        fbytes fdate fduration human2bytes
        hash2attr attr2hash
        filesize md5sum shasum sha1sum sha256sum
        m11r_checkdigit
        parse_credentials
        parsewords
    /, @EXPORT);

use constant HUMAN_SUFFIXES => {
    'B' => 0,
    'K' => 10, 'KB' => 10, 'KIB' => 10,
    'M' => 20, 'MB' => 20, 'MIB' => 20,
    'G' => 30, 'GB' => 30, 'GIB' => 30,
    'T' => 40, 'TB' => 40, 'TIB' => 40,
    'P' => 50, 'PB' => 50, 'PIB' => 50,
    'E' => 60, 'EB' => 60, 'EIB' => 60,
    'Z' => 70, 'ZB' => 70, 'ZIB' => 70,
    'Y' => 80, 'YB' => 80, 'YIB' => 80,
};

sub attr2hash {
    my $t = shift || return ();
    my %as = ();
    for ($t ? (split /\s*[,;]\s*/, $t) : ()) {
        my ($k, $v) = (split /\s*=>?\s*|\s*\:\s*/, $_);
        $as{$k} = $v;
    }
    return %as;
}
sub hash2attr {
    my %h = @_;
    my @r;
    while (my ($k, $v) = each %h) {
        push @r, sprintf("%s=%s", $k, $v // '');
    }
    return "" unless @r;
    return join "; ", @r;
}
sub shasum {
    my $f = shift;
    my $alg = shift || 1;
    my $sha1 = Digest::SHA->new($alg);
    my $sum = '';
    return $sum unless -e $f;
    open( my $sha1_fh, '<', $f) or (carp("Can't open '$f': $!") && return $sum);
    if ($sha1_fh) {
        binmode($sha1_fh);
        $sha1->addfile($sha1_fh);
        $sum = $sha1->hexdigest;
        close($sha1_fh);
    }
    return $sum;
}
sub sha1sum { goto &shasum }
sub sha256sum { &shasum(shift(@_), 256) }
sub md5sum {
    my $f = shift;
    my $md5 = Digest::MD5->new;
    my $sum = '';
    return $sum unless -e $f;
    open( my $md5_fh, '<', $f) or (carp("Can't open '$f': $!") && return $sum);
    if ($md5_fh) {
        binmode($md5_fh);
        $md5->addfile($md5_fh);
        $sum = $md5->hexdigest;
        close($md5_fh);
    }
    return $sum;
}
sub filesize {
    my $f = shift;
    my $filesize = 0;
    $filesize = (stat $f)[7] if -e $f;
    return $filesize;
}
sub m11r_checkdigit {
    my $test = pop;
    croak "Incorrect input digit-string" if !$test || $test =~ m/[^0-9]/g;
    my $len = length($test);
    my $iters = ($len + (($len & 1) ? 1 : 0)) / 2;
    my @digits = split(//, $test); # Get all digits from input string of chars
    #printf "Test=%s; len=%d; iters=%d\n", $test, $len, $iters;

    my $w_lim = 10; # Maximum for round-robin(10) weight list: 1,2,3,4,5,6,7,8,9,10,1,2,3,4,5...
    my $step = 2; # Step for weight list offset calculation for next iteration

    # Calculation sum for one weight list by ofset
    my $calc = sub {
        my $off = shift || 0;
        my $s = 0;
        for (my $i = 0; $i < $len; $i++) {
            my $w = (($i + $off) % $w_lim) + 1;
            $s += ($w * $digits[$i]);
            #printf " > i=%d; d=%d; w=%d; sum=%d\n", $i, $digits[$i], $w, $s;
        }
        return $s % 11;
    };

    # Main cycle
    my $sum = 0;
    for (my $j = 0; $j < $iters; $j++) {
        my $offset = $j*$step;
        $sum = $calc->($offset);
        #printf " >> j=%d; offset=%d; sum=%d\n", $j, $offset, $sum;
        last if $sum < 10;
    }
    $sum = 0 if $sum >= 10; # 0 if incorrect again
    #printf " >>> sum=%d\n", $sum;

    return $sum;
}
sub parsewords {
    my $s = shift;
    my @words = grep { defined && length } quotewords(qr/[\,\;\t ]/, 0, $s);
    return @words;
}
sub parse_credentials {
    my $url = shift || return ();
    my $uri = (ref($url) eq 'URI') ? $url : URI->new($url);
    my $info = $uri->userinfo() // "";
    my $user = $info;
    my $pass = $info;
    $user =~ s/:.*//;
    $pass =~ s/^[^:]*://;
    return (uri_unescape($user // ''), uri_unescape($pass // ''));
}
sub fbytes {
    my $n = int(shift);
    if ($n >= 1024 ** 3) {
        return sprintf "%.3g GB", $n / (1024 ** 3);
    } elsif ($n >= 1024 ** 2) {
        return sprintf "%.3g MB", $n / (1024.0 * 1024);
    } elsif ($n >= 1024) {
        return sprintf "%.3g KB", $n / 1024.0;
    } else {
        return "$n B"; # bytes
    }
}
sub human2bytes {
    my $h = shift || 0;
    return 0 unless $h;
    my ($bts, $sfx) = $h =~ /([0-9.]+)\s*([a-zA-Z]*)/;
    return 0 unless $bts;
    my $exp = HUMAN_SUFFIXES->{($sfx ? uc($sfx) : "B")} || 0;
    return ceil($bts * (2 ** $exp));
}
sub fduration {
    my $msecs = shift || 0;
    my $secs = int($msecs);
    my $hours = int($secs / (60*60));
    $secs -= $hours * 60*60;
    my $mins = int($secs / 60);
    $secs %= 60;
    if ($hours) {
        return sprintf("%d hours %d minutes", $hours, $mins);
    } elsif ($mins >= 2) {
        return sprintf("%d minutes", $mins);
    } elsif ($secs < 2*60) {
        return sprintf("%.4f seconds", $msecs);
    } else {
        $secs += $mins * 60;
        return sprintf("%d seconds", $secs);
    }
}
sub fdate {
    my $d = shift || 0;
    my $g = shift || 0;
    return "unknown" unless $d;
    return dtf(DATETIME_GMT_FORMAT, $d, 1) if $g;
    return dtf(DATETIME_FORMAT . " " . tz_diff(), $d);
}
sub explain {
    require Data::Dumper;
    my $dumper = Data::Dumper->new( [shift] );
    $dumper->Indent(1)->Terse(1);
    $dumper->Sortkeys(1) if $dumper->can("Sortkeys");
    return $dumper->Dump;
}

# Colored helper functions
sub green {  IS_TTY ? colored(['bright_green'],  sprintf(shift, @_)) : sprintf(shift, @_) }
sub red {    IS_TTY ? colored(['bright_red'],    sprintf(shift, @_)) : sprintf(shift, @_) }
sub yellow { IS_TTY ? colored(['bright_yellow'], sprintf(shift, @_)) : sprintf(shift, @_) }
sub cyan {   IS_TTY ? colored(['bright_cyan'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub blue {   IS_TTY ? colored(['bright_blue'],   sprintf(shift, @_)) : sprintf(shift, @_) }
sub magenta {IS_TTY ? colored(['bright_magenta'],sprintf(shift, @_)) : sprintf(shift, @_) }

1;

__END__
