#!/usr/bin/perl -w
#########################################################################
#
# Serz Minus (Sergey Lepenkov), <abalama@cpan.org>
#
# Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved
#
# This is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.
#
#########################################################################
use Test::More tests => 15;
use NoRFC::Number;
use NoRFC::Store;

# Create instance
my $store = NoRFC::Store->new(
    file => "norfc_test.db.tmp",
    attributes => "RaiseError=0; PrintError=0; sqlite_unicode=1",
); #note(explain($store));
ok($store->status, "Create Store instance") or diag($store->error);
ok($store->ping, "Ping") or diag($store->error);

# Set 3 objects
my $norfcnum1 = NoRFC::Number->new('NoRFC-R.2000.001.1.6');
my $norfcnum2 = NoRFC::Number->new('NoRFC-R.2000.001.2.3');
my $norfcnum3 = NoRFC::Number->new('NoRFC-R.2000.001.3.0');
#note $norfcnum3->as_string;

# Add new records
{
    ok($store->add(
        num         => $norfcnum1->as_string,
        pub         => time(),
        category    => "Test1",
        title       => "Test document #1",
        author      => "Anonymous",
        content     => "Test document #1",
        signature   => "none",
    ), "Add new record 1") or diag($store->error);
    ok($store->add(
        num         => $norfcnum2->as_string,
        pub         => time(),
        category    => "Test2",
        title       => "Test document #2",
        author      => "Anonymous",
        content     => "Test document #2",
        signature   => "none",
    ), "Add new record 2") or diag($store->error);
    ok($store->add(
        num         => $norfcnum3->as_string,
        pub         => time(),
        category    => "Test3",
        title       => "Test document #3",
        author      => "Anonymous",
        content     => "Test document #3",
        signature   => "none",
    ), "Add new record 3") or diag($store->error);
}

# Get count
{
    my $count = $store->count();
    is($count, 3, "Count=3") or diag($store->error);
}

# Get record data
my %info = ();
{
    %info = $store->get($norfcnum3->as_string);
    ok($info{id} && $store->status, "Get record data for NoRFC-R.2000.001.3.0") or diag($store->error);
    is($info{content}, "Test document #3", "Check content") or diag($store->error);
}
#note(explain(\%info));

# Update record
{
    ok($store->set(
        num         => $norfcnum3->as_string,
        pub         => time(),
        category    => "Test3",
        title       => "Test document #3 - is updated",
        author      => "Anonymous",
        content     => "Test document #3 - is updated",
        signature   => "none",
    ), "Update record 2") or diag($store->error);
}

# Get record info
{
    my $id = $store->getid($norfcnum3->as_string);
    ok($id, "Get record id") or diag($store->error);
    %info = $store->getinfo($norfcnum3->as_string);
    is($info{id}, $id, "Get record info") or diag($store->error);
}

# Get all records
{
    my @all = $store->getall();
    is(scalar(@all), 3, "Get all records") or diag($store->error);
    #note(explain(\@all));
}

# Add relations
{
    $store->reladd(
        number      => $norfcnum3->as_string,
        external    => $norfcnum1->as_string,
        name        => "Obsoletes",
    ) or die($store->error);
    $store->reladd(
        number      => $norfcnum3->as_string,
        external    => $norfcnum2->as_string,
        name        => "Updates",
    ) or die($store->error);
}

# Get relations
{
    my @relations = $store->relget($norfcnum3->as_string);
    is(scalar(@relations), 2, "Get all relations") or diag($store->error);
    #note(explain(\@relations));
}

# Delete relations
{
    ok($store->reldel($norfcnum3->as_string), "Delete relations") or diag($store->error);
}

# Delete all records
{
    ok($store->truncate(), "Delete all records") or diag($store->error);
}

1;

__END__

prove -lv t/03-store.t
