#!/usr/bin/perl -w
#########################################################################
#
# Serz Minus (Sergey Lepenkov), <abalama@cpan.org>
#
# Copyright (C) 1998-2023 D&D Corporation. All Rights Reserved
#
# This is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.
#
#########################################################################
use Test::More;
use Mojo::File qw/path/;
use CTK::Util qw/dtf bsave/;
use NoRFC::Const;
use NoRFC::Client;

use constant BASE_URLS => [
        'http://localhost:8732/',
        'https://localhost:8732/',
    ];

plan skip_all => "Currently a developer-only test" unless -d '.svn' || -d ".git";

# Create the instance
my $client;
my $base_url;
for (@{(BASE_URLS)}) {
    $base_url = $_;
    $client = NoRFC::Client->new(
        url                 => $base_url,
        insecure            => 1, # IO::Socket::SSL::set_defaults(SSL_verify_mode => 0); # Skip verify for test only!
        max_redirects       => 2,
        connect_timeout     => 3,
        inactivity_timeout  => 5,
        request_timeout     => 10,
        username            => "test",
        password            => "test",
    );
    last if $client->check(); # Check is ok
    note($base_url, ": ", $client->error || 'unknown');
}
plan skip_all => "Can't initialize the client" unless $client->status;
#note $client->trace;
plan skip_all => "Authorization failed" unless $client->res->headers->header('X-Authorized');

# Start testing
plan tests => 16;
note $base_url;

# General info (GET method)
{
    my $status = $client->info();
    ok($status, "Get server info")
        or diag($client->res->json("/message") || $client->res->message);
    #note $status ? explain($client->res->json("/categories")) : $client->trace;
    note $client->trace unless $status;
}

# GET gendocnum
my $docnum = "";
{
    my $status = $client->gendocnum();
    ok($status, "GET gendocnum")
        or diag($client->res->json("/message") || $client->res->message);
    $docnum = $client->res->json("/number");
    note $status ? $docnum : $client->trace;
}

# Publish document
my $source_file = "fake06.tmp";
{
    my $savestatus = bsave( $source_file, "Content of test document\n" );
    ok($savestatus, "Save test source file") or diag(sprintf("Can't save file \"%s\"", $source_file));
    my $status = $client->publish(
            number      => $docnum,
            published   => time(),
            category    => "Test",
            title       => "Test document",
            author      => "Anonymous",
            attributes  => "foo=bar",
            fileraw     => $source_file,
            #filesig     => $file_sig,
        );
    ok($status, "Publish document")
        or diag($client->res->json("/message") || $client->res->message);
    note $client->trace unless $status;
}

# Search documents
{
    my $status = $client->search(
            text    => $docnum,
            category=> undef,
            page    => 1,
            limit   => 3,
        );
    ok($status, "Search documents")
        or diag($client->res->json("/message") || $client->res->message);
    is($client->res->json("/count"), 1, "Found one document") or diag($client->trace);
}

# File info
{
    my $status = $client->info($docnum);
    ok($status, "Get file info")
        or diag($client->res->json("/message") || $client->res->message);
    ok($client->res->json("/id"), "Get id of doc") or diag($client->trace);
}

# Download file from server
my $document_file = sprintf("%s.tmp", $docnum);
{
    my $cli_status = $client->download(sprintf("%s.txt", $docnum) => $document_file);
    ok($cli_status, "Download file")
        or diag($client->res->json("/message") || $client->res->message);
    note $client->trace unless $cli_status;
}

# Add relation
my $extnum = "NoRFC-R202301755";
{
    my $status = $client->reladd($docnum, $extnum, "Updates");
    ok($status, "Add relation")
        or diag($client->res->json("/message") || $client->res->message);
    ok($client->res->json("/id"), "Get id of relation") or diag($client->trace);
}

# Get relation
{
    my $status = $client->relget($docnum, $extnum);
    ok($status, "Get relation")
        or diag($client->res->json("/message") || $client->res->message);
    ok($client->res->json("/status"), "Get id of doc by relation") or diag($client->trace);
}

# Delete relation
{
    my $status = $client->reldel($docnum, $extnum);
    ok($status, "Delete relation")
        or diag($client->res->json("/message") || $client->res->message);
    ok($client->res->json("/status"), "Deleted relation") or diag($client->trace);
}

# Delete document
{
    my $status = $client->remove($docnum);
    ok($status, "Delete document")
        or diag($client->res->json("/message") || $client->res->message);
    note $client->trace unless $client->res->json("/status");
}

1;

__END__

prove -lv t/06-client.t
